from __future__ import annotations
from abc import ABC
import inspect
from dataclasses import astuple, dataclass, field, fields
from pathlib import Path
import struct
import traceback
import typing
from typing import Union, get_origin, get_args
import zlib

from fprime_gds.common.fpy.types import (
    SPECIFIC_FLOAT_TYPES,
    SPECIFIC_INTEGER_TYPES,
    MACROS,
    MAX_DIRECTIVE_SIZE,
    MAX_DIRECTIVES_COUNT,
    SPECIFIC_NUMERIC_TYPES,
    SIGNED_INTEGER_TYPES,
    UNSIGNED_INTEGER_TYPES,
    CompileState,
    FieldReference,
    FppTypeClass,
    FpyCallable,
    FpyCmd,
    FpyMacro,
    FpyReference,
    FpyScope,
    FpyTypeCtor,
    FpyVariable,
    InternalIntType,
    InternalStringType,
    NothingType,
    TopDownVisitor,
    Visitor,
    create_scope,
    get_ref_fpp_type_class,
    is_instance_compat,
)

from fprime_gds.common.fpy.error import CompileError
# In Python 3.10+, the `|` operator creates a `types.UnionType`.
# We need to handle this for forward compatibility, but it won't exist in 3.9.
try:
    from types import UnionType

    UNION_TYPES = (Union, UnionType)
except ImportError:
    UNION_TYPES = (Union,)

from fprime_gds.common.fpy.bytecode.directives import (
    BINARY_STACK_OPS,
    BOOLEAN_OPERATORS,
    NUMERIC_OPERATORS,
    UNARY_STACK_OPS,
    AllocateDirective,
    BinaryStackOp,
    ConstCmdDirective,
    FloatMultiplyDirective,
    FloatTruncateDirective,
    IntMultiplyDirective,
    MemCompareDirective,
    NoOpDirective,
    StackOpDirective,
    IntegerTruncate64To16Directive,
    IntegerTruncate64To32Directive,
    IntegerTruncate64To8Directive,
    FloatLogDirective,
    IntegerSignedExtend16To64Directive,
    IntegerSignedExtend32To64Directive,
    IntegerSignedExtend8To64Directive,
    StackCmdDirective,
    StorePrmDirective,
    IntegerZeroExtend16To64Directive,
    IntegerZeroExtend32To64Directive,
    IntegerZeroExtend8To64Directive,
    Directive,
    FloatExtendDirective,
    ExitDirective,
    LoadDirective,
    StoreTlmValDirective,
    GotoDirective,
    IfDirective,
    NotDirective,
    PushValDirective,
    SignedIntToFloatDirective,
    StoreDirective,
    UnaryStackOp,
    UnsignedIntToFloatDirective,
    WaitAbsDirective,
    WaitRelDirective,
)
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.event_json_loader import EventJsonLoader
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.templates.prm_template import PrmTemplate
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.type_exceptions import TypeException
from fprime.common.models.serialize.numerical_types import (
    U32Type,
    U16Type,
    U64Type,
    U8Type,
    I64Type,
    F32Type,
    F64Type,
    FloatType,
    IntegerType,
    NumericalType,
)
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.bool_type import BoolType
from fprime_gds.common.fpy.parser import (
    AstBinaryOp,
    AstBoolean,
    AstElif,
    AstElifs,
    AstExpr,
    AstGetAttr,
    AstGetItem,
    AstNumber,
    AstOp,
    AstReference,
    AstScopedBody,
    AstString,
    Ast,
    AstBody,
    AstLiteral,
    AstIf,
    AstAssign,
    AstFuncCall,
    AstUnaryOp,
    AstVar,
)
from fprime.common.models.serialize.type_base import BaseType as FppType


class AssignIds(TopDownVisitor):
    """assigns a unique id to each node to allow it to be indexed in a dict"""

    def __init__(self):
        self.next_id = 0

    def visit_default(self, node, state):
        node.id = self.next_id
        self.next_id += 1


class CreateVariables(Visitor):
    """finds all variable declarations and adds them to the variable scope"""

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        existing = state.variables.get(node.variable.var, None)
        if not existing:
            # idk what this var is. make sure it's a valid declaration
            if node.var_type is None:
                # error because this isn't an annotated assignment. right now all declarations must be annotated
                state.err(
                    "Must provide a type annotation for new variables", node.variable
                )
                return

            var = FpyVariable(node.var_type, node)
            # new var. put it in the table under this scope
            state.variables[node.variable.var] = var
            state.runtime_values[node.variable.var] = var

        if existing and node.var_type is not None:
            # redeclaring an existing variable
            state.err(f"{node.variable.var} already declared", node)
            return


class ResolveReferences(TopDownVisitor):
    """for each reference, resolve it in a specific scope based on its
    syntactic position, or fail if could not resolve"""

    def is_type_constant_size(self, type: FppTypeClass) -> bool:
        """return true if the type is statically sized"""
        if issubclass(type, StringType):
            return False

        if issubclass(type, ArrayType):
            return self.is_type_constant_size(type.MEMBER_TYPE)

        if issubclass(type, SerializableType):
            for _, arg_type, _, _ in type.MEMBER_LIST:
                if not self.is_type_constant_size(arg_type):
                    return False
            return True

        return True

    def get_attr_of_ref(
        self, parent: FpyReference, node: AstGetAttr, state: CompileState
    ) -> FpyReference | None:
        """resolve a GetAttr node relative to a given FpyReference. return the
        resolved ref, or None if none could be found. Will raise errors if not found"""

        if isinstance(parent, (FpyCallable, type)):
            # right now we don't support resolving something after a callable/type
            state.err("Invalid syntax", node)
            return None

        if isinstance(parent, dict):
            # parent is a scope
            attr = parent.get(node.attr, None)
            if attr is None:
                state.err("Unknown attribute", node)
                return None
            return attr

        # parent is a ch, prm, const, or field

        value_type = get_ref_fpp_type_class(parent)

        assert value_type != NothingType

        if not issubclass(value_type, (SerializableType, TimeType)):
            # trying to do arr.x, but arr is not a struct
            state.err(
                "Invalid syntax (tried to access named member of a non-struct type)",
                node,
            )
            return None

        if not self.is_type_constant_size(value_type):
            state.err(
                f"{value_type} has non-constant sized members, cannot access members",
                node,
            )
            return None

        member_list: list[tuple[str, FppTypeClass]] = None
        if issubclass(value_type, SerializableType):
            member_list = [t[0:2] for t in value_type.MEMBER_LIST]
        else:
            # if it is a time type, there are some "implied" members
            member_list = []
            member_list.append(("time_base", U16Type))
            member_list.append(("time_context", U8Type))
            member_list.append(("seconds", U32Type))
            member_list.append(("useconds", U32Type))

        offset = 0
        for arg_name, arg_type in member_list:
            if arg_name == node.attr:
                return FieldReference(parent, arg_type, offset, name=arg_name)
            offset += arg_type.getMaxSize()

        state.err(f"Unknown member {node.attr}", node)
        return None

    def get_item_of_ref(
        self, parent: FpyReference, node: AstGetItem, state: CompileState
    ) -> FpyReference | None:
        """resolve a GetItem node relative to a given FpyReference. return the
        resolved ref, or None if none could be found. Will raise errors if not found"""

        if isinstance(parent, (FpyCallable, type, dict)):
            # right now we don't support resolving index after a callable/type/scope
            state.err("Invalid syntax", node)
            return None

        # parent is a ch, prm, const, or field

        value_type = get_ref_fpp_type_class(parent)

        assert value_type != NothingType

        if not issubclass(value_type, ArrayType):
            # trying to do struct[0], but struct is not an array
            state.err(
                "Invalid syntax (tried to access indexed member of a non-array type)",
                node.item,
            )
            return None

        if not self.is_type_constant_size(value_type):
            state.err(
                f"{value_type.__name__} has non-constant sized members, cannot access members",
                node,
            )
            return None

        offset = 0
        for i in range(0, value_type.LENGTH):
            if i == node.item.value:
                return FieldReference(parent, value_type.MEMBER_TYPE, offset, idx=i)
            offset += value_type.MEMBER_TYPE.getMaxSize()

        state.err(
            f"Array access out-of-bounds (access: {node.item}, array size: {value_type.LENGTH})",
            node.item,
        )
        return None

    def resolve_if_ref(self, node: AstExpr, ns: FpyScope, state: CompileState) -> bool:
        """if the node is a reference, try to resolve it in the given scope, and return true if success.
        otherwise, if it is not a reference, return true as it doesn't need to be resolved
        """
        if not is_instance_compat(node, AstReference):
            return True

        return self.resolve_ref_in_ns(node, ns, state) is not None

    def resolve_ref_in_ns(
        self, node: AstReference, ns: FpyScope, state: CompileState
    ) -> FpyReference | None:
        """recursively resolves a reference in a scope, returning the resolved ref
        or none if none could be found."""
        if isinstance(node, AstVar):
            if not isinstance(ns, dict):
                return None
            ref = ns.get(node.var, None)
            if ref is None:
                return None
            state.resolved_references[node] = ref
            return ref

        parent = self.resolve_ref_in_ns(node.parent, ns, state)
        if parent is None:
            # couldn't resolve parent
            return None

        if isinstance(node, AstGetItem):
            ref = self.get_item_of_ref(parent, node, state)
            state.resolved_references[node] = ref
            return ref

        assert isinstance(node, AstGetAttr)
        ref = self.get_attr_of_ref(parent, node, state)
        state.resolved_references[node] = ref
        return ref

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        # function refs must be callables
        if not self.resolve_ref_in_ns(node.func, state.callables, state):
            state.err("Unknown callable", node.func)
            return

        for arg in node.args if node.args is not None else []:
            # arg value refs must have values at runtime
            if not self.resolve_if_ref(arg, state.runtime_values, state):
                state.err("Unknown runtime value", arg)
                return

    def visit_AstIf_AstElif(self, node: Union[AstIf, AstElif], state: CompileState):
        # if condition expr refs must be "runtime values" (tlm/prm/const/etc)
        if not self.resolve_if_ref(node.condition, state.runtime_values, state):
            state.err("Unknown runtime value", node.condition)
            return

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        # lhs/rhs side of stack op, if they are refs, must be refs to "runtime vals"
        if not self.resolve_if_ref(node.lhs, state.runtime_values, state):
            state.err("Unknown runtime value", node.lhs)
            return
        if not self.resolve_if_ref(node.rhs, state.runtime_values, state):
            state.err("Unknown runtime value", node.rhs)
            return

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        if not self.resolve_if_ref(node.val, state.runtime_values, state):
            state.err("Unknown runtime value", node.val)
            return

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var = self.resolve_ref_in_ns(node.variable, state.variables, state)
        if not var:
            state.err("Unknown variable", node.variable)
            return

        if node.var_type is not None:
            type = self.resolve_ref_in_ns(node.var_type, state.types, state)
            if not type:
                state.err("Unknown type", node.var_type)
                return
            var.type = type

        if not self.resolve_if_ref(node.value, state.runtime_values, state):
            state.err("Unknown runtime value", node.value)
            return

    def visit_AstReference(self, node: AstReference, state: CompileState):
        # make sure that all refs are resolved when we get to them
        if node not in state.resolved_references:
            state.err("Unknown variable", node)
            return


class CheckUseBeforeDeclare(Visitor):

    def __init__(self):
        self.currently_declared_vars: list[FpyVariable] = []

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var = state.resolved_references[node.variable]

        if var.declaration != node:
            # this is not the node that declares this variable
            return

        # this node declares this variable

        self.currently_declared_vars.append(var)

    def visit_AstReference(self, node: AstReference, state: CompileState):
        ref = state.resolved_references[node]
        if not isinstance(ref, FpyVariable):
            return

        if ref.declaration.variable == node:
            # this is the initial name of the variable. don't crash
            return

        if ref not in self.currently_declared_vars:
            state.err("Variable used before declared", node)
            return


class PickAndConvertTypes(Visitor):
    """stores in state the fprime type of each expression, or NothingType if the expr had no type"""

    def coerce_expr_type(
        self, node: AstExpr, type: FppTypeClass, state: CompileState
    ) -> bool:
        node_type = state.expr_types[node]
        if self.can_coerce_type(node_type, type):
            state.type_coercions[node] = type
            return True
        state.err(f"Expected {type.__name__}, found {node_type.__name__}", node)
        return False

    def can_coerce_type(self, type: FppTypeClass, to_type: FppTypeClass) -> bool:
        if type == to_type:
            return True
        if issubclass(type, IntegerType) and issubclass(to_type, NumericalType):
            # we can coerce any integer into any other number
            return True
        if issubclass(type, FloatType) and issubclass(to_type, FloatType):
            # we can convert any float into any float
            return True
        if type == InternalStringType and issubclass(to_type, StringType):
            # we can convert the internal String type to any string type
            return True

        return False

    def pick_intermediate_type(
        self, arg_types: list[FppTypeClass], op: BinaryStackOp | UnaryStackOp
    ) -> FppTypeClass:

        if op in BOOLEAN_OPERATORS:
            return BoolType

        non_numeric = any(not issubclass(t, NumericalType) for t in arg_types)

        if op == BinaryStackOp.EQUAL or op == BinaryStackOp.NOT_EQUAL:
            if non_numeric:
                if len(set(arg_types)) != 1:
                    # can only compare equality between the same types
                    return None
                return arg_types[0]

        # all arguments should be numeric
        if non_numeric:
            # cannot find intermediate type
            return None

        if op == BinaryStackOp.DIVIDE or op == BinaryStackOp.EXPONENT:
            # always do true division over floats, python style
            return F64Type

        float = any(issubclass(t, FloatType) for t in arg_types)
        unsigned = any(t in UNSIGNED_INTEGER_TYPES for t in arg_types)

        if float:
            # at least one arg is a float
            return F64Type

        if unsigned:
            # at least one arg is unsigned
            return U64Type

        return I64Type

    def visit_AstNumber(self, node: AstNumber, state: CompileState):
        # give a best guess as to the final type of this node. we don't actually know
        # its bitwidth or signedness yet
        if isinstance(node.value, float):
            result_type = F64Type
        else:
            result_type = InternalIntType
        state.expr_types[node] = result_type

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        lhs_type = state.expr_types[node.lhs]
        rhs_type = state.expr_types[node.rhs]

        intermediate_type = self.pick_intermediate_type([lhs_type, rhs_type], node.op)
        if intermediate_type is None:
            state.err(
                f"Op {node.op} undefined for {lhs_type.__name__}, {rhs_type.__name__}",
                node,
            )
            return

        if not self.coerce_expr_type(node.lhs, intermediate_type, state):
            return
        if not self.coerce_expr_type(node.rhs, intermediate_type, state):
            return

        # okay now find which actual directive we're going to use based on this intermediate
        # type, and save it

        dir = None
        if (
            node.op == BinaryStackOp.EQUAL or node.op == BinaryStackOp.NOT_EQUAL
        ) and intermediate_type not in SPECIFIC_NUMERIC_TYPES:
            dir = MemCompareDirective
        else:
            dir = BINARY_STACK_OPS[node.op][intermediate_type]

        state.stack_op_directives[node] = dir

        result_type = None
        if node.op in NUMERIC_OPERATORS:
            result_type = intermediate_type
        else:
            result_type = BoolType
        state.expr_types[node] = result_type

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        val_type = state.expr_types[node.val]

        intermediate_type = self.pick_intermediate_type([val_type], node.op)
        if intermediate_type is None:
            state.err(f"Op {node.op} undefined for {val_type.__name__}", node)
            return

        if not self.coerce_expr_type(node.val, intermediate_type, state):
            return

        # okay now find which actual directive we're going to use based on this intermediate
        # type, and save it

        chosen_dir = UNARY_STACK_OPS[node.op][intermediate_type]

        result_type = None
        if node.op in NUMERIC_OPERATORS:
            result_type = intermediate_type
        else:
            result_type = BoolType

        state.stack_op_directives[node] = chosen_dir
        state.expr_types[node] = result_type

    def visit_AstString(self, node: AstString, state: CompileState):
        state.expr_types[node] = InternalStringType

    def visit_AstBoolean(self, node: AstBoolean, state: CompileState):
        state.expr_types[node] = BoolType

    def visit_AstReference(self, node: AstReference, state: CompileState):
        ref = state.resolved_references[node]
        state.expr_types[node] = get_ref_fpp_type_class(ref)
        if isinstance(node, AstGetItem):
            # the node of the index number has no expression value, it's an arg
            # but only at syntax level
            state.expr_types[node.item] = NothingType

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_references[node.func]
        assert isinstance(func, FpyCallable)
        func_args = func.args
        node_args = node.args if node.args else []

        if len(node_args) < len(func_args):
            state.errors.append(
                CompileError(
                    f"Missing arguments (expected {len(func_args)} found {len(node_args)})",
                    node,
                )
            )
            return
        if len(node_args) > len(func_args):
            state.errors.append(
                CompileError(
                    f"Too many arguments (expected {len(func_args)} found {len(node_args)})",
                    node,
                )
            )
            return

        for value_expr, arg in zip(node_args, func_args):
            arg_name, arg_type = arg

            if not self.coerce_expr_type(value_expr, arg_type, state):
                return

        # got thru all args successfully
        state.expr_types[node] = func.return_type

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var_type = state.resolved_references[node.variable].type

        if not self.coerce_expr_type(node.value, var_type, state):
            return

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not is_instance_compat(node, AstExpr), node


class AllocateVariables(Visitor):
    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        existing_var = state.resolved_references[node.variable]

        assert existing_var is not None
        assert existing_var.type is not None

        value_size = existing_var.type.getMaxSize()

        if existing_var.lvar_offset is None:
            # doesn't have an lvar idx, allocate one
            lvar_offset = state.lvar_array_size_bytes
            state.lvar_array_size_bytes += value_size
            existing_var.lvar_offset = lvar_offset


class CalculateConstExprValues(Visitor):
    """for each expr, try to calculate its constant value and store it in a map. stores None if no value could be
    calculated at compile time, and NothingType if the expr had no value"""

    def const_coerce_type(self, from_val: FppType, to_type: FppTypeClass) -> FppType:
        if type(from_val) == to_type:
            return from_val
        if issubclass(to_type, StringType):
            assert type(from_val) == InternalStringType, type(from_val)
            return to_type(from_val.val)
        if issubclass(to_type, FloatType):
            assert issubclass(type(from_val), NumericalType), type(from_val)
            return to_type(float(from_val.val))
        if issubclass(to_type, IntegerType):
            assert issubclass(type(from_val), IntegerType), type(from_val)
            return to_type(int(from_val.val))
        assert False, (from_val, type(from_val), to_type)

    def visit_AstLiteral(self, node: AstLiteral, state: CompileState):
        literal_type = state.expr_types[node]

        if literal_type == NothingType:
            value = NothingType()
        else:
            try:
                value = literal_type(node.value)
            except TypeException as e:
                state.err(f"For type {literal_type.__name__}: {e}", node)
                return

        coerced_type = state.type_coercions.get(node, None)
        if coerced_type is not None:
            try:
                value = self.const_coerce_type(value, coerced_type)
            except TypeException as e:
                state.err(f"For type {coerced_type.__name__}: {e}", node)
                return
        state.expr_values[node] = value

    def visit_AstReference(self, node: AstReference, state: CompileState):
        ref = state.resolved_references[node]
        expr_type = state.expr_types[node]

        if isinstance(ref, (ChTemplate, PrmTemplate, FpyVariable)):
            # we do not try to calculate or predict these values at compile time
            expr_value = None
        elif isinstance(ref, FieldReference):
            if isinstance(ref.parent, FppType):
                # ref to a field of a constant
                # get the field
                expr_value = ref.get_from(ref.parent)
            else:
                # ref to a field of smth else. no runtime val
                expr_value = None
        elif isinstance(ref, FppType):
            # constant value
            expr_value = ref
        elif isinstance(ref, FpyCallable):
            # a reference to a callable doesn't have a value, you have to actually
            # call the func
            expr_value = NothingType()
        elif isinstance(ref, type):
            # a reference to a type doesn't have a value, and so doesn't have a type,
            # in and of itself. if this were a function call to the type's ctor then
            # it would have a value
            expr_value = NothingType()
        elif isinstance(ref, dict):
            # a ref to a scope doesn't have a value
            expr_value = NothingType()
        else:
            assert False, ref

        if expr_value is None:
            # cannot calculate at compile time
            state.expr_values[node] = None
            return

        assert isinstance(expr_value, expr_type), (expr_value, expr_type)

        coerced_type = state.type_coercions.get(node, None)
        if coerced_type is not None:
            try:
                expr_value = self.const_coerce_type(expr_value, coerced_type)
            except TypeException as e:
                state.err(f"For type {expr_type.__name__}: {e}", node)
                return
        state.expr_values[node] = expr_value

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        func = state.resolved_references[node.func]
        assert isinstance(func, FpyCallable)
        # gather arg values
        arg_values = [
            state.expr_values[e] for e in (node.args if node.args is not None else [])
        ]
        unknown_value = any(v for v in arg_values if v is None)
        if unknown_value:
            # we will have to calculate this at runtime
            state.expr_values[node] = None
            return

        expr_value = None

        if isinstance(func, FpyTypeCtor):
            # actually construct the type
            if issubclass(func.type, SerializableType):
                instance = func.type()
                # pass in args as a dict
                # t[0] is the arg name
                arg_dict = {t[0]: v for t, v in zip(func.type.MEMBER_LIST, arg_values)}
                instance._val = arg_dict
                expr_value = instance

            elif issubclass(func.type, ArrayType):
                instance = func.type()
                instance._val = arg_values
                expr_value = instance

            elif func.type == TimeType:
                expr_value = TimeType(*arg_values)

            else:
                # no other FppTypeClasses have ctors
                assert False, func.return_type
        else:
            # don't try to calculate the value of this function call
            # it's something like a cmd or macro
            state.expr_values[node] = None
            return

        assert isinstance(expr_value, func.return_type), (expr_value, func.return_type)

        coerced_type = state.type_coercions.get(node, None)
        if coerced_type is not None:
            try:
                expr_value = self.const_coerce_type(expr_value, coerced_type)
            except TypeException as e:
                state.err(f"For type {func.return_type.__name__}: {e}", node)
                return
        state.expr_values[node] = expr_value

    def visit_AstOp(self, node: AstOp, state: CompileState):
        # we do not calculate compile time value of operators at the moment
        state.expr_values[node] = None

    def visit_default(self, node, state):
        # coding error, missed an expr
        assert not is_instance_compat(node, AstExpr), node


class GenerateConstExprDirectives(Visitor):
    """for each expr with a constant compile time value, generate
    directives for how to put it in its register"""

    def visit_AstExpr(self, node: AstExpr, state: CompileState):
        if node in state.directives:
            # already have directives associated with this node
            return

        expr_value = state.expr_values[node]

        if expr_value is None:
            # no const value
            return

        if isinstance(expr_value, NothingType):
            # nothing type has no value
            state.directives[node] = []
            return

        # it has a constant value at compile time
        serialized_expr_value = expr_value.serialize()

        # push it to the stack
        state.directives[node] = [PushValDirective(serialized_expr_value)]


class GenerateExprMacrosAndCmds(Visitor):
    """for each expr whose value is not known at compile time, but can be calculated at run time,
    generate directives to calculate the value and put it in its register. for each command
    or macro, generate directives for calling them with appropriate arg values"""

    def get_64_bit_type(self, type: FppTypeClass) -> FppTypeClass:
        assert type in SPECIFIC_NUMERIC_TYPES, type
        return (
            I64Type
            if type in SIGNED_INTEGER_TYPES
            else U64Type if type in UNSIGNED_INTEGER_TYPES else F64Type
        )

    def truncate_from_64_bits(
        self, from_type: FppTypeClass, new_size: int
    ) -> list[Directive]:

        assert new_size in (1, 2, 4, 8), new_size
        assert from_type.getMaxSize() == 8, from_type.getMaxSize()

        if new_size == 8:
            # already correct size
            return []

        if from_type == F64Type:
            # only one option for float trunc
            assert new_size == 4, new_size
            return [FloatTruncateDirective()]

        # must be an int
        assert issubclass(from_type, IntegerType), from_type

        if new_size == 1:
            return [IntegerTruncate64To8Directive()]
        elif new_size == 2:
            return [IntegerTruncate64To16Directive()]

        return [IntegerTruncate64To32Directive()]

    def extend_to_64_bits(self, type: FppTypeClass) -> list[Directive]:
        if type.getMaxSize() == 8:
            # already 8 bytes
            return []
        if type == F32Type:
            return [FloatExtendDirective()]

        # must be an int
        assert issubclass(type, IntegerType), type

        from_size = type.getMaxSize()
        assert from_size in (1, 2, 4, 8), from_size

        if type in SIGNED_INTEGER_TYPES:
            if from_size == 1:
                return [IntegerSignedExtend8To64Directive()]
            elif from_size == 2:
                return [IntegerSignedExtend16To64Directive()]
            else:
                return [IntegerSignedExtend32To64Directive()]
        else:
            if from_size == 1:
                return [IntegerZeroExtend8To64Directive()]
            elif from_size == 2:
                return [IntegerZeroExtend16To64Directive()]
            else:
                return [IntegerZeroExtend32To64Directive()]

    def convert_type(
        self, from_type: FppTypeClass, to_type: FppTypeClass
    ) -> list[Directive]:
        if from_type == to_type:
            return []

        # only valid runtime type conversion is between two numeric types
        assert (
            from_type in SPECIFIC_NUMERIC_TYPES and to_type in SPECIFIC_NUMERIC_TYPES
        ), (
            from_type,
            to_type,
        )
        # also invalid to convert from a float to an integer at runtime due to loss of precision
        assert not (
            from_type in SPECIFIC_FLOAT_TYPES and to_type in SPECIFIC_INTEGER_TYPES
        ), (
            from_type,
            to_type,
        )

        dirs = []
        # first go to 64 bit width
        dirs.extend(self.extend_to_64_bits(from_type))
        from_64_bit = self.get_64_bit_type(from_type)
        to_64_bit = self.get_64_bit_type(to_type)

        # now convert from int to float if necessary
        if from_64_bit == U64Type and to_64_bit == F64Type:
            dirs.append(UnsignedIntToFloatDirective())
            from_64_bit = F64Type
        elif from_64_bit == I64Type and to_64_bit == F64Type:
            dirs.append(SignedIntToFloatDirective())
            from_64_bit = F64Type
        elif from_64_bit == U64Type or from_64_bit == I64Type:
            assert to_64_bit == U64Type or to_64_bit == I64Type
            # conversion from signed to unsigned int is implicit, doesn't need code gen
            from_64_bit = to_64_bit

        assert from_64_bit == to_64_bit, (from_64_bit, to_64_bit)

        # now truncate back down to desired size
        dirs.extend(self.truncate_from_64_bits(to_64_bit, to_type.getMaxSize()))
        return dirs

    def visit_AstReference(self, node: AstReference, state: CompileState):
        if node in state.directives:
            # already know how to put it on stack, or it is impossible
            return

        expr_type = state.expr_types[node]
        ref = state.resolved_references[node]

        directives = []

        # does not have a constant compile time value

        # first, put it in an lvar. then load it from the lvar onto stack

        # the offset of the field in the parent type
        offset_in_parent_val = 0
        # the offset of the lvar the parent type is stored in
        offset_in_lvar_array = 0

        base_ref = ref

        # if it's a field ref, find the parent and the offset in the parent
        while isinstance(base_ref, FieldReference):
            offset_in_parent_val += base_ref.offset
            base_ref = base_ref.parent

        if isinstance(base_ref, ChTemplate):
            # put it in an lvar
            offset_in_lvar_array = state.lvar_array_size_bytes
            state.lvar_array_size_bytes += base_ref.get_type_obj().getMaxSize()
            directives.append(
                StoreTlmValDirective(base_ref.get_id(), offset_in_lvar_array)
            )
        elif isinstance(base_ref, PrmTemplate):
            # put it in an lvar
            offset_in_lvar_array = state.lvar_array_size_bytes
            state.lvar_array_size_bytes += base_ref.get_type_obj().getMaxSize()
            directives.append(
                StorePrmDirective(base_ref.get_id(), offset_in_lvar_array)
            )
        elif isinstance(base_ref, FpyVariable):
            # already should be in an lvar
            offset_in_lvar_array = base_ref.lvar_offset
        else:
            assert (
                False
            ), base_ref  # ref should either be impossible to put on stack or should have a compile time val

        # load from the lvar
        directives.append(
            LoadDirective(
                offset_in_lvar_array + offset_in_parent_val, expr_type.getMaxSize()
            )
        )
        converted_type = state.type_coercions.get(node, None)
        if converted_type is not None:
            directives.extend(self.convert_type(expr_type, converted_type))

        state.directives[node] = directives

    def visit_AstBinaryOp(self, node: AstBinaryOp, state: CompileState):
        if node in state.directives:
            # already know how to put it on stack
            return

        directives = []

        expr_type = state.expr_types[node]

        lhs_dirs = state.directives[node.lhs]
        rhs_dirs = state.directives[node.rhs]

        # which variant of the op did we pick?
        dir = state.stack_op_directives[node]

        # generate the actual op itself
        directives: list[Directive] = lhs_dirs + rhs_dirs
        if dir == MemCompareDirective:
            lhs_type = state.type_coercions.get(node.lhs, None)
            if lhs_type is None:
                lhs_type = state.expr_types[node.lhs]
            rhs_type = state.type_coercions.get(node.rhs, None)
            if rhs_type is None:
                rhs_type = state.expr_types[node.rhs]
            assert lhs_type == rhs_type, (lhs_type, rhs_type)
            directives.append(dir(lhs_type.getMaxSize()))
            if node.op == BinaryStackOp.NOT_EQUAL:
                directives.append(NotDirective())
        elif dir == NoOpDirective:
            # don't include no op
            pass
        else:
            directives.append(dir())

        # and convert the result of the op into the desired result of this expr
        converted_type = state.type_coercions.get(node, None)
        if converted_type is not None:
            directives.extend(self.convert_type(expr_type, converted_type))

        state.directives[node] = directives

    def visit_AstUnaryOp(self, node: AstUnaryOp, state: CompileState):
        if node in state.directives:
            # already know how to put it on stack
            return

        directives = []

        expr_type = state.expr_types[node]

        val_dirs = state.directives[node.val]

        # which variant of the op did we pick?
        dir = state.stack_op_directives[node]
        # generate the actual op itself
        directives: list[Directive] = val_dirs

        if node.op == UnaryStackOp.NEGATE:
            # in this case, we also need to push -1
            if dir == FloatMultiplyDirective:
                directives.append(PushValDirective(F64Type(-1).serialize()))
            elif dir == IntMultiplyDirective:
                directives.append(PushValDirective(I64Type(-1).serialize()))

        directives.append(dir())
        # and convert the result of the op into the desired result of this expr
        converted_type = state.type_coercions.get(node, None)
        if converted_type is not None:
            directives.extend(self.convert_type(expr_type, converted_type))

        state.directives[node] = directives

    def visit_AstFuncCall(self, node: AstFuncCall, state: CompileState):
        node_args = node.args if node.args is not None else []
        func = state.resolved_references[node.func]
        dirs = state.directives.get(node, [])
        if len(dirs) > 0:
            # already know how to put this on the stack
            return
        if isinstance(func, FpyCmd):
            const_args = not any(
                state.expr_values[arg_node] is None for arg_node in node_args
            )
            if const_args:
                # can just hardcode this cmd
                arg_bytes = bytes()
                for arg_node in node_args:
                    arg_value = state.expr_values[arg_node]
                    arg_bytes += arg_value.serialize()
                dirs = [ConstCmdDirective(func.cmd.get_op_code(), arg_bytes)]
            else:
                arg_byte_count = 0
                # push all args to the stack
                # keep track of how many bytes total we have pushed
                for arg_node in node_args:
                    node_dirs = state.directives[arg_node]
                    assert len(node_dirs) >= 1
                    dirs.extend(node_dirs)
                    converted_type = state.type_coercions.get(arg_node, None)
                    if converted_type is None:
                        converted_type = state.expr_types[arg_node]
                    arg_byte_count += converted_type.getMaxSize()
                # then push cmd opcode to stack as u32
                dirs.append(
                    PushValDirective(U32Type(func.cmd.get_op_code()).serialize())
                )
                # now that all args are pushed to the stack, pop them and opcode off the stack
                # as a command
                dirs.append(StackCmdDirective(arg_byte_count))
        elif isinstance(func, FpyMacro):
            # put all arg values on stack
            for arg_node in node_args:
                node_dirs = state.directives[arg_node]
                assert len(node_dirs) >= 1
                dirs.extend(node_dirs)

            dirs.append(func.dir())
        else:
            dirs = None

        # perform type conversion if called for
        coerced_type = state.type_coercions.get(node, None)
        if coerced_type is not None:
            dirs.extend(self.convert_type(func.return_type, coerced_type))
        state.directives[node] = dirs

    def visit_AstAssign(self, node: AstAssign, state: CompileState):
        var = state.resolved_references[node.variable]
        state.directives[node] = state.directives[node.value] + [
            StoreDirective(var.lvar_offset, var.type.getMaxSize())
        ]


class CountNodeDirectives(Visitor):
    """count the number of directives that will be generated by each node"""

    def visit_AstIf(self, node: AstIf, state: CompileState):
        count = 0
        # include the condition
        count += state.node_dir_counts[node.condition]
        # include if stmt
        count += 1
        # include body
        count += state.node_dir_counts[node.body]
        # include a goto end of if
        count += 1

        if node.elifs is not None:
            count += state.node_dir_counts[node.elifs]
        if node.els is not None:
            count += state.node_dir_counts[node.els]

        state.node_dir_counts[node] = count

    def visit_AstElifs(self, node: AstElifs, state: CompileState):
        count = 0
        for case in node.cases:
            count += state.node_dir_counts[case]

        state.node_dir_counts[node] = count

    def visit_AstElif(self, node: AstElif, state: CompileState):
        count = 0
        # include the condition
        count += state.node_dir_counts[node.condition]
        # include if stmt
        count += 1
        # include body
        count += state.node_dir_counts[node.body]
        # include a goto end of if
        count += 1

        state.node_dir_counts[node] = count

    def visit_AstBody(self, node: Union[AstBody, AstScopedBody], state: CompileState):
        count = 0
        if isinstance(node, AstScopedBody):
            # add one for lvar array alloc
            count += 1
        for stmt in node.stmts:
            count += state.node_dir_counts[stmt]

        state.node_dir_counts[node] = count

    def visit_default(self, node, state):
        state.node_dir_counts[node] = (
            len(state.directives[node]) if state.directives.get(node) is not None else 0
        )


class CalculateStartLineIdx(TopDownVisitor):
    """based on the number of directives generated by each node, calculate the start line idx
    of each node's directives"""

    def visit_AstBody(self, node: Union[AstBody, AstScopedBody], state: CompileState):
        if node not in state.start_line_idx:
            state.start_line_idx[node] = 0

        start_idx = state.start_line_idx[node]

        line_idx = start_idx
        if isinstance(node, AstScopedBody):
            # include lvar alloc
            line_idx += 1

        for stmt in node.stmts:
            state.start_line_idx[stmt] = line_idx
            line_idx += state.node_dir_counts[stmt]

    def visit_AstIf(self, node: AstIf, state: CompileState):
        line_idx = state.start_line_idx[node]
        state.start_line_idx[node.condition] = line_idx
        line_idx += state.node_dir_counts[node.condition]
        # include if stmt
        line_idx += 1
        state.start_line_idx[node.body] = line_idx
        line_idx += state.node_dir_counts[node.body]
        # include goto stmt
        line_idx += 1
        if node.elifs is not None:
            state.start_line_idx[node.elifs] = line_idx
            line_idx += state.node_dir_counts[node.elifs]
        if node.els is not None:
            state.start_line_idx[node.els] = line_idx
            line_idx += state.node_dir_counts[node.els]

    def visit_AstElifs(self, node: AstElifs, state: CompileState):
        line_idx = state.start_line_idx[node]
        for case in node.cases:
            state.start_line_idx[case] = line_idx
            line_idx += state.node_dir_counts[case]

    def visit_AstElif(self, node: AstElif, state: CompileState):
        line_idx = state.start_line_idx[node]
        state.start_line_idx[node.condition] = line_idx
        line_idx += state.node_dir_counts[node.condition]
        # include if dir
        line_idx += 1
        state.start_line_idx[node.body] = line_idx
        line_idx += state.node_dir_counts[node.body]
        # include a goto end of if
        line_idx += 1


class GenerateBodyDirectives(Visitor):
    """concatenate all directives together for each AstBody"""

    def visit_AstIf(self, node: AstIf, state: CompileState):
        start_line_idx = state.start_line_idx[node]

        all_dirs = []

        cases: list[tuple[AstExpr, AstBody]] = []
        goto_ends: list[GotoDirective] = []

        cases.append((node.condition, node.body))

        if node.elifs is not None:
            for case in node.elifs.cases:
                cases.append((case.condition, case.body))

        for case in cases:
            case_dirs = []
            # put the conditional on top of stack
            case_dirs.extend(state.directives[case[0]])
            # include if stmt (update the end idx later)
            if_dir = IfDirective(-1)

            case_dirs.append(if_dir)
            # include body
            case_dirs.extend(state.directives[case[1]])
            # include a temporary goto end of if, will be refined later
            goto_dir = GotoDirective(-1)
            case_dirs.append(goto_dir)
            goto_ends.append(goto_dir)

            # if false, skip the body and goto
            if_dir.false_goto_dir_index = (
                start_line_idx + len(all_dirs) + len(case_dirs)
            )

            all_dirs.extend(case_dirs)

        if node.els is not None:
            all_dirs.extend(state.directives[node.els])

        for goto in goto_ends:
            goto.dir_idx = start_line_idx + len(all_dirs)

        state.directives[node] = all_dirs

    def visit_AstBody(self, node: Union[AstBody, AstScopedBody], state: CompileState):
        dirs = []
        if isinstance(node, AstScopedBody):
            dirs.append(AllocateDirective(state.lvar_array_size_bytes))
        for stmt in node.stmts:
            stmt_dirs = state.directives.get(stmt, None)
            if stmt_dirs is not None:
                dirs.extend(stmt_dirs)

        state.directives[node] = dirs

def get_base_compile_state(dictionary: str) -> CompileState:
    """return the initial state of the compiler, based on the given dict path"""
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    prm_json_dict_loader = PrmJsonLoader(dictionary)
    (prm_id_dict, prm_name_dict, versions) = prm_json_dict_loader.construct_dicts(
        dictionary
    )
    event_json_dict_loader = EventJsonLoader(dictionary)
    (event_id_dict, event_name_dict, versions) = event_json_dict_loader.construct_dicts(
        dictionary
    )
    # the type name dict is a mapping of a fully qualified name to an fprime type
    # here we put into it all types found while parsing all cmds, params and tlm channels
    type_name_dict: dict[str, FppTypeClass] = cmd_json_dict_loader.parsed_types
    type_name_dict.update(ch_json_dict_loader.parsed_types)
    type_name_dict.update(prm_json_dict_loader.parsed_types)
    type_name_dict.update(event_json_dict_loader.parsed_types)

    # enum const dict is a dict of fully qualified enum const name (like Ref.Choice.ONE) to its fprime value
    enum_const_name_dict: dict[str, FppType] = {}

    # find each enum type, and put each of its values in the enum const dict
    for name, typ in type_name_dict.items():
        if issubclass(typ, EnumType):
            for enum_const_name, val in typ.ENUM_DICT.items():
                enum_const_name_dict[name + "." + enum_const_name] = typ(
                    enum_const_name
                )

    # insert the implicit types into the dict
    type_name_dict["Fw.Time"] = TimeType
    for typ in SPECIFIC_NUMERIC_TYPES:
        type_name_dict[typ.get_canonical_name()] = typ
    type_name_dict["bool"] = BoolType
    # note no string type at the moment

    cmd_response_type = type_name_dict["Fw.CmdResponse"]
    callable_name_dict: dict[str, FpyCallable] = {}
    # add all cmds to the callable dict
    for name, cmd in cmd_name_dict.items():
        cmd: CmdTemplate
        args = []
        for arg_name, _, arg_type in cmd.arguments:
            args.append((arg_name, arg_type))
        # cmds are thought of as callables with a Fw.CmdResponse return value
        callable_name_dict[name] = FpyCmd(cmd_response_type, args, cmd)

    # for each type in the dict, if it has a constructor, create an FpyTypeCtor
    # object to track the constructor and put it in the callable name dict
    for name, typ in type_name_dict.items():
        args = []
        if issubclass(typ, SerializableType):
            for arg_name, arg_type, _, _ in typ.MEMBER_LIST:
                args.append((arg_name, arg_type))
        elif issubclass(typ, ArrayType):
            for i in range(0, typ.LENGTH):
                args.append(("e" + str(i), typ.MEMBER_TYPE))
        elif issubclass(typ, TimeType):
            args.append(("time_base", U16Type))
            args.append(("time_context", U8Type))
            args.append(("seconds", U32Type))
            args.append(("useconds", U32Type))
        else:
            # bool, enum, string or numeric type
            # none of these have callable ctors
            continue

        callable_name_dict[name] = FpyTypeCtor(typ, args, typ)

    # for each macro function, add it to the callable dict
    for macro_name, macro in MACROS.items():
        callable_name_dict[macro_name] = macro

    state = CompileState(
        tlms=create_scope(ch_name_dict),
        prms=create_scope(prm_name_dict),
        types=create_scope(type_name_dict),
        callables=create_scope(callable_name_dict),
        consts=create_scope(enum_const_name_dict),
    )
    return state


def compile(body: AstScopedBody, dictionary: str) -> list[Directive]:
    state = get_base_compile_state(dictionary)
    passes: list[Visitor] = [
        AssignIds(),
        # based on assignment syntax nodes, we know which variables exist where
        CreateVariables(),
        # now that variables have been defined, all names/attributes/indices (references)
        # should be defined
        ResolveReferences(),
        CheckUseBeforeDeclare(),
        # now that we know what all refs point to, we should be able to figure out the type
        # of every expression
        PickAndConvertTypes(),
        # now that expr types have been narrowed down, we can allocate lvar space for variables
        AllocateVariables(),
        # okay, now that we're sure we're passing in all the right args to each func,
        # we can calculate values of type ctors etc etc
        CalculateConstExprValues(),
        # for expressions which have constant values, generate corresponding directives
        # to put the expr on the stack
        GenerateConstExprDirectives(),
        # generate directives to calculate exprs, macros and cmds at runtime and put them
        # on the stack
        GenerateExprMacrosAndCmds(),
        # count the number of directives generated by each node
        CountNodeDirectives(),
        # calculate the index that the node will correspond to in the output file
        CalculateStartLineIdx(),
        # generate directives for each body node, including the root
        GenerateBodyDirectives(),
    ]

    for compile_pass in passes:
        compile_pass.run(body, state)
        for error in state.errors:
            print(error)
            exit(1)

    dirs = state.directives[body]
    if len(dirs) > MAX_DIRECTIVES_COUNT:
        print(CompileError(f"Too many directives in sequence (expected less than {MAX_DIRECTIVES_COUNT}, had {len(dirs)})"))
        exit(1)

    return dirs



"""
@brief Utility class to read config files and provide configuration values

After the first instance of this class is initialized, any class can obtain
the same instance by calling the static getInstance function. This allows any
part of the python program retrieve configuration data.

Based on the ConfigManager class written by Len Reder in the fprime Gse

@author R. Joseph Paetz

@date Created July 25, 2018

@license Copyright 2018, California Institute of Technology.
         ALL RIGHTS RESERVED. U.S. Government Sponsorship acknowledged.
"""
import configparser

# Custom type modules
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)


class ConfigBadTypeException(Exception):
    def __init__(self, config_name, type_str):
        """
        Constructor

        Args:
            config_name (string): Name of the config containing the bad type
            type_str (string): Bad type string that caused the error
        """
        super().__init__(f"Invalid type string {type_str} read in configuration {config_name}")


class ConfigManager(configparser.ConfigParser):
    """
    This class provides a single entrypoint for all configurable properties,
    """

    __instance = None
    __prop = None

    def __init__(self):
        """
        Constructor

        Creates a ConfigManager object with the default configuration values

        Returns:
            An instance of the ConfigManager class. Default configurations
            will be used until the set_configs method is called!
        """
        # Cannot use super() function since ConfigParser is an old-style class
        configparser.ConfigParser.__init__(self)

        # Set default properties
        self.__prop = {}
        self._set_defaults()
        self.file_path = None

    def set_configs(self, f):
        """
        Sets the configuration values to those in the given file

        Will raise an exception if the file does not exist.

        Args:
            f (string): Path to a file object to read
        """
        self.file_path = f
        self.read_file(open(f))

    @staticmethod
    def get_instance():
        """
        Return instance of singleton.

        Returns:
            The current ConfigManager object for this python application
        """
        if ConfigManager.__instance is None:
            ConfigManager.__instance = ConfigManager()
        return ConfigManager.__instance

    def get_type(self, name):
        """
        Retrieve a type from the config file for parsing

        It is assumed the setting is in the types section

        Args:
            name (string): Name of the type to retrieve

        Returns:
            If the name is valid, returns an object of a type derived from
            TypeBase. Otherwise, raises ConfigBadTypeException
        """
        type_str = self.get("types", name)

        if type_str == "U8":
            return U8Type()
        if type_str == "U16":
            return U16Type()
        if type_str == "U32":
            return U32Type()
        if type_str == "U64":
            return U64Type()
        if type_str == "I8":
            return I8Type()
        if type_str == "I16":
            return I16Type()
        if type_str == "I32":
            return I32Type()
        if type_str == "I64":
            return I64Type()
        if type_str == "F32":
            return F32Type()
        if type_str == "F64":
            return F64Type()
        # These are types for parsing, so they need to be number types
        # Other types can be added later
        raise ConfigBadTypeException(name, type_str)

    def get_file_path(self):
        """
        Return file loaded for this configuration

        :return: file path
        """
        return self.file_path

    def _set_defaults(self):
        """
        Used by the constructor to set all ConfigParser defaults

        Establishes a dictionary of sections and then a dictionary of keyword,
        value association for each section.
        """

        ########################## TYPES ###########################
        # These configs give the types of fields in the binary data

        self.__prop["types"] = {
            "msg_len": "U32",
            "FwPacketDescriptorType": "U32",
            "FwChanIdType": "U32",
            "FwEventIdType": "U32",
            "FwOpcodeType": "U32",
            "FwTlmPacketizeIdType": "U16",
            "key_val": "U16",
        }
        self._set_section_defaults("types")

        ######################### COLORS ###########################
        # Colors are hex codes in BGR format

        self.__prop["colors"] = {
            "warning_lo": "0x00BCED",
            "warning_hi": "0x0073E5",
            "fatal": "0x0000FF",
            "command": "0xFF0000",
            "red": "0x0000FF",
            "orange": "0x0073E5",
            "yellow": "0x00BCED",
        }
        self._set_section_defaults("colors")

        self.__prop["framing"] = {"use_key": "False", "key_val": "0x0"}
        self._set_section_defaults("framing")

    def _set_section_defaults(self, section):
        """
        For a section set up the default values.

        Args:
            section: Section to set all the defaults config values for
        """
        self.add_section(section)
        for (key, value) in self.__prop[section].items():
            self.set(section, key, str(value))



"""
fw_type_json_loader.py:

Loads flight dictionary (JSON) and returns name based Python dictionaries of Fw types

@author jawest
"""

from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException

class FwTypeJsonLoader(JsonLoader):
    """Class to load python based Fw type dictionaries"""

    TYPE_DEFINITIONS_FIELD = "typeDefinitions"

    def construct_dicts(self, _):
        """
        Constructs and returns python dictionaries keyed on id and name

        Args:
            _: Unused argument (inherited)
        Returns:
            A tuple with two Fw type dictionaries (python type dict):
            (id_dict, name_dict). The keys should be the type id and
            name fields respectively and the values should be type name
            strings. Note: An empty id dictionary is returned since there 
            are no id fields in the Fw type alias JSON dictionary entries.
        """
        id_dict = {}
        name_dict = {}

        if self.TYPE_DEFINITIONS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.TYPE_DEFINITIONS_FIELD}' field: {str(self.json_file)}"
            )

        for type_def in self.json_dict[self.TYPE_DEFINITIONS_FIELD]:
            try:
                if type_def["kind"] == "alias":
                    name = str(type_def["qualifiedName"])
                    # Only consider names with the pattern Fw*Type
                    if name.startswith("Fw") and name.endswith("Type"):
                        name_dict[type_def["qualifiedName"]] = type_def["underlyingType"]["name"]
            except KeyError as e:
                raise GdsDictionaryParsingException(
                    f"{str(e)} key missing from Type Definition dictionary entry: {str(type_def)}"
                )

        return (
            dict(sorted(id_dict.items())),
            dict(sorted(name_dict.items())),
            self.get_versions(),
        )



"""fprime_gds.plugin.system: implementation of plugins

This file contains the implementation and registration of plugins for fprime_gds. Primarily, it defines the Plugins
class that handles plugins. Users can acquire the Plugin singleton with `Plugin.system()`.

This file also imports and registers plugin implementations built-into fprime-gds. These plugins are not registered
using entrypoints.

@author lestarch
"""
import copy
import os
import importlib
import inspect
import logging
from typing import Iterable, List, Union

import pluggy

from fprime_gds.plugin.definitions import Plugin, PluginType, PROJECT_NAME


# Handy constants
LOGGER = logging.getLogger(__name__)


class PluginException(Exception):
    pass


class InvalidCategoryException(PluginException):
    pass


class PluginsNotLoadedException(PluginException):
    pass


class Plugins(object):
    """GDS plugin system providing a plugin Singleton for use across the GDS

    GDS plugins are broken into categories (e.g. framing) that represent the key features users can adjust. Each GDS
    application will support and load the plugins for a given category.
    """

    PLUGIN_ENVIRONMENT_VARIABLE = "FPRIME_GDS_EXTRA_PLUGINS"
    PLUGIN_METADATA = None
    _singleton = None

    def __init__(self, categories: Union[None, List] = None):
        """Initialize the plugin system with specific categories

        Initialize the plugin system with support for the supplied categories. Only plugins for the specified categories
        will be loaded for use. Other plugins will not be available for use.

        Args:
            categories: None for all categories otherwise a list of categories
        """
        self.metadata = copy.deepcopy(self.get_plugin_metadata())
        categories = self.get_all_categories() if categories is None else categories
        self.categories = categories
        self.manager = pluggy.PluginManager(PROJECT_NAME)

        # Load hook specifications from only the configured categories
        for category in categories:
            self.manager.add_hookspecs(self.metadata[category]["class"])

        # Load plugins from setuptools entrypoints and the built-in plugins (limited to category)
        try:
            self.manager.load_setuptools_entrypoints(PROJECT_NAME)
        except Exception as e:
            LOGGER.warning("Failed to load entrypoint plugins: %s", e)
        # Load plugins from environment variable specified modules
        for token in [
            token
            for token in os.environ.get(self.PLUGIN_ENVIRONMENT_VARIABLE, "").split(";")
            if token
        ]:
            module, class_token = token.split(":")
            try:
                imported_module = importlib.import_module(module)
                module_class = (
                    module
                    if class_token == ""
                    else getattr(imported_module, class_token, imported_module)
                )
                self.register_plugin(module_class)
            except ImportError as imp:
                LOGGER.debug("Failed to load %s.%s as plugin", module, class_token)

        # Load built-in plugins
        for category in categories:
            for built_in in self.metadata[category]["built-in"]:
                self.register_plugin(built_in)

    def get_plugins(self, category) -> Iterable:
        """Get available plugins for the given category

        Gets all plugin implementors of "category" by looking for register_<category>_plugin implementors. If such a
        function does not exist then this results in an exception.

        Args:
            category: category of the plugin requested

        Return:
            validated list of plugin implementor classes
        """
        try:
            plugin_classes = getattr(self.manager.hook, f"register_{category}_plugin")()
        except KeyError as error:
            raise InvalidCategoryException(f"Invalid plugin category: {error}")

        return [
            Plugin(category, self.get_category_plugin_type(category), plugin_class)
            for plugin_class in plugin_classes
            if self.validate_selection(category, plugin_class)
        ]

    def start_loading(self, category: str):
        """Start a category loading

        When loading plugins via the CLI, it is imperative to distinguish between the no loaded implementors case, and
        the case where loading was never attempted. This sets the variable in metadata to [] to indicate the loading
        was attempted.
        """
        metadata = self.metadata[category]
        metadata["bound_classes"] = (
            metadata["bound_classes"] if "bound_classes" in metadata else []
        )

    def add_bound_class(self, category: str, bound_class: List[object]):
        """Add class for plugin category with constructor arguments bound

        Called from the plugin cli parser, this will add a class ready for zero-argument construction to the categories'
        list of classes. This is the plugin class with constructor arguments bound to the cli arguments that fill those
        values. For SELECTION plugins, only a single instance is allowed. For FEATURE plugins, multiple bound classes
        are  allowed.

        Args:
            category: category to set
            bound_class: constructor argument bound class
        """
        self.start_loading(category)
        metadata = self.metadata[category]
        metadata["bound_classes"].append(bound_class)
        assert (
            metadata["type"] == PluginType.FEATURE
            or len(metadata["bound_classes"]) == 1
        ), f"Multiple selections for: {category}"

    def get_selected_class(self, category: str) -> object:
        """Get the selected constructor-bound class for the category"""
        metadata = self.metadata[category]
        assert (
            metadata["type"] == PluginType.SELECTION
        ), "Features allow multiple plugins"
        try:
            return metadata["bound_classes"][0]
        except (KeyError, IndexError):
            raise PluginsNotLoadedException(
                f"Plugins not loaded for category: {category}"
            )

    def get_feature_classes(self, category: str) -> object:
        """Get the selected instance for the category"""
        metadata = self.metadata[category]
        assert (
            metadata["type"] == PluginType.FEATURE
        ), "Selections have single instances"
        try:
            return metadata["bound_classes"]
        except KeyError:
            raise PluginsNotLoadedException(
                f"Plugins not loaded for category: {category}"
            )

    def register_plugin(self, module_or_class):
        """Register a plugin directly

        Allows local registration of plugin implementations that are shipped as part of the GDS package.

        Args:
            module_or_class: module or class that has plugin implementations
        """
        self.manager.register(module_or_class)

    def get_categories(self):
        """Get plugin categories"""
        return self.categories

    @classmethod
    def get_all_categories(cls):
        """Get all plugin categories"""
        return cls.get_plugin_metadata().keys()

    @classmethod
    def get_plugin_metadata(cls, category: str = None):
        """Get the metadata describing a given category

        F Prime supports certain plugin types that break down into categories. Each category has a set of built-in
        plugins class type, and plugin type. This function will load that metadata for the supplied category. If no
        category is supplied then the complete metadata block is returned.

        On first invocation, the method will load plugin types and construct the metadata object.

        Warning:
            Since the loaded plugins may use features of the plugin system, these plugins must be imported at call time
            instead of imported at the top of the module to prevent circular imports.

        Return:
            plugin metadata definitions
        """
        # Load on the first time only
        if cls.PLUGIN_METADATA is None:
            from fprime_gds.executables.apps import GdsFunction, GdsApp
            from fprime_gds.common.handlers import DataHandlerPlugin
            from fprime_gds.common.communication.framing import (
                FramerDeframer,
                FpFramerDeframer,
            )
            from fprime_gds.common.communication.ccsds.chain import SpacePacketSpaceDataLinkFramerDeframer
            from fprime_gds.common.communication.adapters.base import (
                BaseAdapter,
                NoneAdapter,
            )
            from fprime_gds.common.communication.adapters.ip import IpAdapter
            from fprime_gds.executables.apps import CustomDataHandlers

            try:
                from fprime_gds.common.communication.adapters.uart import SerialAdapter
            except ImportError:
                SerialAdapter = None
            cls.PLUGIN_METADATA = {
                "framing": {
                    "class": FramerDeframer,
                    "type": PluginType.SELECTION,
                    "built-in": [FpFramerDeframer, SpacePacketSpaceDataLinkFramerDeframer],
                },
                "communication": {
                    "class": BaseAdapter,
                    "type": PluginType.SELECTION,
                    "built-in": [
                        adapter
                        for adapter in [NoneAdapter, IpAdapter, SerialAdapter]
                        if adapter is not None
                    ],
                },
                "gds_function": {
                    "class": GdsFunction,
                    "type": PluginType.FEATURE,
                    "built-in": [],
                },
                "gds_app": {
                    "class": GdsApp,
                    "type": PluginType.FEATURE,
                    "built-in": [CustomDataHandlers],
                },
                "data_handler": {
                    "class": DataHandlerPlugin,
                    "type": PluginType.FEATURE,
                    "built-in": [],
                },
            }
            assert cls.PLUGIN_METADATA is not None, "Failed to set plugin metadata"
        return (
            cls.PLUGIN_METADATA[category]
            if category is not None
            else cls.PLUGIN_METADATA
        )

    @classmethod
    def get_category_plugin_type(cls, category):
        """Get the plugin type given the category"""
        return cls.get_plugin_metadata(category)["type"]

    @classmethod
    def get_category_specification_class(cls, category):
        """Get the plugin class given the category"""
        return cls.get_plugin_metadata(category)["class"]

    @classmethod
    def validate_selection(cls, category, result):
        """Validate the result of plugin hook

        Validates the result of a plugin hook call to ensure the result meets the expected properties for plugins of the
        given category. Primarily this ensures that this plugin returns a concrete subclass of the expected type.

        Args:
            category: category of plugin used
            result: result from the plugin hook call
        Return:
            True when the plugin passes validation, False otherwise
        """
        # Typing library not intended for introspection at runtime, thus we maintain a map of plugin specification
        # functions to the types expected as a return value. When this is not found, plugins may continue without
        # automatic validation.
        try:
            expected_class = cls.get_category_specification_class(category)
            # Validate the result
            if not issubclass(result, expected_class):
                LOGGER.warning(
                    f"{result.__name__} is not a subclass of {expected_class.__name__}. Not registering."
                )
                return False
            elif inspect.isabstract(result):
                LOGGER.warning(
                    f"{result.__name__} is an abstract class. Not registering."
                )
                return False
        except KeyError:
            LOGGER.warning(
                f"Plugin not registered for validation. Continuing without validation."
            )
        return True

    @classmethod
    def system(cls, categories: Union[None, List] = None) -> "Plugins":
        """Get plugin system singleton

        Constructs the plugin system singleton (when it has yet to be constructed) then returns the singleton. The
        singleton will support specific categories and further requests for a singleton will cause an assertion error
        unless the categories match or is None.

        Args:
            categories: a list of categories to support or None to use the existing categories

        Returns:
            plugin system
        """
        # Singleton undefined, construct it
        if cls._singleton is None:
            cls._singleton = cls(
                cls.get_all_categories() if categories is None else categories
            )
        # Ensure categories was unspecified or matches the singleton
        assert (
            categories is None or cls._singleton.categories == categories
        ), f"Inconsistent plugin categories: {categories} vs {cls._singleton.categories}"
        return cls._singleton



from pathlib import Path
import tempfile
import traceback
from fprime_gds.common.fpy.types import deserialize_directives
from fprime_gds.common.fpy.model import DirectiveErrorCode, FpySequencerModel
from fprime_gds.common.fpy.bytecode.directives import Directive
from fprime_gds.common.fpy.main import assemble_main, compile_main, disassemble_main
from fprime_gds.common.loaders.ch_json_loader import ChJsonLoader
from fprime_gds.common.loaders.cmd_json_loader import CmdJsonLoader
from fprime_gds.common.loaders.event_json_loader import EventJsonLoader
from fprime_gds.common.loaders.prm_json_loader import PrmJsonLoader
from fprime_gds.common.testing_fw.api import IntegrationTestAPI

default_dictionary = str(
    Path(__file__).parent.parent.parent.parent.parent
    / "test"
    / "fprime_gds"
    / "common"
    / "fpy"
    / "RefTopologyDictionary.json"
)


def compile_seq(fprime_test_api, seq: str) -> list[Directive]:
    input_file = tempfile.NamedTemporaryFile(suffix=".fpy", delete=False)
    output_file = tempfile.NamedTemporaryFile(suffix=".bin", delete=False)
    Path(input_file.name).write_text(seq)
    compile_main(["-d", default_dictionary, "-o", output_file.name, input_file.name])

    # also, run some additional tests: try reading the bin file, turning it into assembly,
    # parsing the assembly, writing it to disk and making sure it's the same as the bin file

    # okay write the fpybc to file
    bytecode_file = tempfile.NamedTemporaryFile(suffix=".fpybc", delete=False)
    disassemble_main([output_file.name, "-o", bytecode_file.name])
    assembled_file = tempfile.NamedTemporaryFile(suffix=".bin", delete=False)
    assemble_main([bytecode_file.name, "-o", assembled_file.name])
    # okay now check that the assembled file is the same as the compiled file
    assert Path(assembled_file.name).read_bytes() == Path(output_file.name).read_bytes()

    return output_file.name


def lookup_type(fprime_test_api, type_name: str):
    dictionary = default_dictionary  # fprime_test_api.pipeline.dictionary_path
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    prm_json_dict_loader = PrmJsonLoader(dictionary)
    (prm_id_dict, prm_name_dict, versions) = prm_json_dict_loader.construct_dicts(
        dictionary
    )
    event_json_dict_loader = EventJsonLoader(dictionary)
    (event_id_dict, event_name_dict, versions) = event_json_dict_loader.construct_dicts(
        dictionary
    )
    type_name_dict = cmd_json_dict_loader.parsed_types
    type_name_dict.update(ch_json_dict_loader.parsed_types)
    type_name_dict.update(prm_json_dict_loader.parsed_types)
    type_name_dict.update(event_json_dict_loader.parsed_types)

    return type_name_dict[type_name]


def run_seq(
    fprime_test_api: IntegrationTestAPI,
    file_name: str,
    tlm: dict[str, bytes] = None,
):
    if tlm is None:
        tlm = {}

    # fprime_test_api.send_and_assert_command("Ref.cmdSeq.RUN", [file.name, "BLOCK"], timeout=4)
    # return

    dictionary = default_dictionary  # fprime_test_api.pipeline.dictionary_path

    deserialized_dirs = deserialize_directives(Path(file_name).read_bytes())

    ch_json_dict_loader = ChJsonLoader(dictionary)
    (ch_id_dict, ch_name_dict, versions) = ch_json_dict_loader.construct_dicts(
        dictionary
    )
    cmd_json_dict_loader = CmdJsonLoader(dictionary)
    (cmd_id_dict, cmd_name_dict, versions) = cmd_json_dict_loader.construct_dicts(
        dictionary
    )
    model = FpySequencerModel(cmd_dict=cmd_id_dict)
    tlm_db = {}
    for chan_name, val in tlm.items():
        ch_template = ch_name_dict[chan_name]
        tlm_db[ch_template.get_id()] = val
    ret = model.run(deserialized_dirs, tlm_db)
    if ret != DirectiveErrorCode.NO_ERROR:
        raise RuntimeError("Sequence returned", ret)


def assert_compile_success(fprime_test_api, seq: str):
    compile_seq(fprime_test_api, seq)


def assert_run_success(fprime_test_api, seq: str, tlm: dict[str, bytes] = None):
    compiled_file = compile_seq(fprime_test_api, seq)

    run_seq(fprime_test_api, compiled_file, tlm)


def assert_compile_failure(fprime_test_api, seq: str):
    try:
        compile_seq(fprime_test_api, seq)
    except BaseException as e:
        traceback.print_exc()
        return
    raise RuntimeError("compile_seq succeeded")


def assert_run_failure(fprime_test_api, seq: str):
    compiled_file = compile_seq(fprime_test_api, seq)
    try:
        run_seq(fprime_test_api, compiled_file)
    except BaseException as e:
        print(e)
        return
    raise RuntimeError("run_seq succeeded")



"""
Handles executing the "command-send" CLI command for the GDS
"""

import difflib
from typing import Iterable, List

from fprime.common.models.serialize.type_exceptions import NotInitializedException

import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime_gds.common.gds_cli.base_commands import BaseCommand
from fprime_gds.common.models.dictionaries import Dictionaries
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.testing_fw import predicates
from fprime_gds.common.testing_fw.api import IntegrationTestAPI


class CommandSendCommand(BaseCommand):
    """
    The implementation for sending a command via the GDS to the spacecraft
    """

    ####################################################################
    #   Utility functions
    ####################################################################
    @staticmethod
    def _get_closest_commands(
        project_dictionary: Dictionaries, command_name: str, num: int = 3
    ) -> List[str]:
        """
        Searches for the closest matching known command(s) to the given command
        name.

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to search for
        :param num: The maximum number of near-matches to return
        :return: A list of the closest matching commands (potentially empty)
        """
        known_commands = project_dictionary.command_name.keys()
        closest_matches = difflib.get_close_matches(command_name, known_commands, n=num)
        return closest_matches

    @staticmethod
    def _get_command_template(
        project_dictionary: Dictionaries, command_name: str
    ) -> CmdTemplate:
        """
        Retrieves the command template for the given command name

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to return a
            template for
        :return: The CmdTemplate object for the given command
        """
        return project_dictionary.command_name[command_name]

    @staticmethod
    def _get_command_help_message(
        project_dictionary: Dictionaries, command_name: str
    ) -> str:
        """
        Returns a string showing a help message for the given GDS command.

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to return a
            help message for
        :return: A help string for the command
        """
        command_template = CommandSendCommand._get_command_template(
            project_dictionary, command_name
        )
        return CommandSendCommand._get_item_string(command_template)

    ####################################################################
    #   Abstract method implementations
    ####################################################################
    @classmethod
    def _get_item_list(
        cls,
        project_dictionary: Dictionaries,
        filter_predicate: predicates.predicate,
    ) -> Iterable[CmdTemplate]:
        """
        Gets a list of available commands in the system and return them in an
        ID-sorted list.

        :param project_dictionary: The dictionary object for the project
            containing the command definitions
        :param filter_predicate: Test API predicate used to filter shown
            channels
        """

        # NOTE: Trying to create a blank CmdData causes errors, so currently
        # just using templates (i.e. this function does nothing)
        def create_empty_command(cmd_template):
            return cmd_template

        command_list = test_api_utils.get_item_list(
            item_dictionary=project_dictionary.command_id,
            search_filter=filter_predicate,
            template_to_data=create_empty_command,
        )
        return command_list

    @classmethod
    def _get_item_string(
        cls,
        item: CmdTemplate,
        json: bool = False,
    ) -> str:
        """
        Converts the given command template into a human-readable string.

        :param item: The CmdTemplate to convert to a string
        :param json: Whether or not to return a JSON representation of "temp"
        :return: A readable string version of "item"
        """
        if not item:
            return ""
        if json:
            # This is questionable whether it should be supported
            from fprime_gds.flask.json import getter_based_json
            return str(getter_based_json(item))

        cmd_string = "%s (%d) | Takes %d arguments.\n" % (
            item.get_full_name(),
            item.get_id(),
            len(item.get_args()),
        )

        cmd_description = item.get_description()
        if cmd_description:
            cmd_string += f"Description: {(cmd_description)}\n" 

        for arg in item.get_args():
            arg_name, arg_description, arg_type = arg
            if not arg_description:
                arg_description = "--no description--"
            # Can't compare against actual module, since EnumType is a serializable
            # type from the dictionary
            if type(arg_type).__name__ == "EnumType":
                arg_description = f"{str(arg_type.keys())} | {arg_description} "

            cmd_string += f"\t{arg_name} ({arg_type.__name__}): {arg_description}\n"

        return cmd_string


    @classmethod
    def _execute_command(cls, args, api: IntegrationTestAPI):
        """
        Logic for sending a command
        """
        command = args.command_name
        arguments = [] if args.arguments is None else args.arguments
        try:
            api.send_command(command, arguments)
        except KeyError:
            cls._log(f"{command} is not a known command")
            close_matches = cls._get_closest_commands(
                api.pipeline.dictionaries, command
            )
            if close_matches:
                cls._log(f"Similar known commands: {close_matches}")
        except NotInitializedException:
            temp = cls._get_command_template(api.pipeline.dictionaries, command)
            cls._log(
                "'%s' requires %d arguments (%d given)"
                % (command, len(temp.get_args()), len(arguments))
            )
            cls._log(cls._get_command_help_message(api.pipeline.dictionaries, command))



"""
string_util.py
Utility functions to process strings to be used in FPrime GDS
@Created March 18, 2021
@`janamian`

Note: This function has an identical copy in fprime-gds
"""

import logging
import re
from typing import Any, Union

LOGGER = logging.getLogger("string_util_logger")


def format_string_template(template: str, value: Union[tuple, list, Any]) -> str:
    """
    Function to format a string template with values. This function is a simple wrapper around the
    format function. It accepts a tuple, list, or single value and passes it to the format function

    Args:
        template (str): String template to be formatted
        value (Union[tuple, list, Any]): Value(s) to be inserted into the template

    Returns:
        str: Formatted string
    """
    if not isinstance(value, (tuple, list)):
        value = (value,)
    try:
        return template.format(*value)
    except (IndexError, ValueError) as e:
        LOGGER.error(
            f"Error formatting string template: {template} with value: {str(value)}"
        )
        raise e


def preprocess_fpp_format_str(format_str: str) -> str:
    """Preprocess a FPP-style format string and convert it to Python format string
    FPP format strings are documented https://nasa.github.io/fpp/fpp-spec.html#Format-Strings
    For example "{x}" -> "{:x}" or "{.2f}" -> "{:.2f}"

    Args:
        format_str (str): FPP-style format string

    Returns:
        str: Python-style format string
    """
    pattern = r"{(\d*\.?\d*[cdxoefgCDXOEFG])}"
    return re.sub(pattern, r"{:\1}", format_str)


def preprocess_c_style_format_str(format_str: str) -> str:
    r"""
    Function to convert C-string style to python format
    without using python interpolation
    Considered the following format for C-string:
    %[flags][width][.precision][length]type

    0- %:                (?<!%)(?:%%)*%
    1- flags:            ([\-\+0\ \#])?
    2- width:            (\d+|\*)?
    3- .precision:       (\.\*|\.\d+)?
    4- length:          `([hLIw]|l{1,2}|I32|I64)?`
    5- conversion_type: `([cCdiouxXeEfgGaAnpsSZ])`

    Note:
    This function will keep the flags, width, and .precision of C-string
    template.

    It will keep f, x, o, and e flags and remove all other types.
    Other types will be duck-typed by python interpreter.

    lengths will also be removed since they are not meaningful to Python interpreter.
    `See: https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting`
    `Regex Source: https://www.regexlib.com/REDetails.aspx?regexp_id=3363`

    For example "%x" -> "{:x}" or "%.2f" -> "{:.2f}"

    Args:
        format_str (str): C-style format string

    Returns:
        str: Python-style format string
    """

    def convert(match_obj: re.Match):
        if match_obj.group() is None:
            return match_obj
        flags, width, precision, _, conversion_type = match_obj.groups()
        format_template = ""
        if flags:
            format_template += f"{flags}"
        if width:
            format_template += f"{width}"
        if precision:
            format_template += f"{precision}"

        if conversion_type and str(conversion_type).lower() in {"f", "x", "o", "e"}:
            format_template += f"{conversion_type}"

        return "{}" if format_template == "" else "{:" + format_template + "}"

    pattern = r"(?<!%)(?:%%)*%([\-\+0\ \#])?(\d+|\*)?(\.\*|\.\d+)?([hLIw]|l{1,2}|I32|I64)?([cCdiouxXeEfgGaAnpsSZ])"  # NOSONAR

    match = re.compile(pattern)

    return re.sub(match, convert, format_str).replace("%%", "%")



"""
ip.py:

Module containing a comm-layer adapter for the Tcp/UDP/IP stack. This pairs with the F prime component "SocketIpDriver"
in order to read data being sent via Tcp or Udp. This is the default adapter used by the system to handle data sent
across a Tcp and/or UDP network interface.

@author lestarch
"""
import abc
import atexit
import logging
import queue
import socket
import threading
import time

import fprime_gds.common.communication.adapters.base
import fprime_gds.common.logger

from fprime_gds.plugin.definitions import gds_plugin_implementation

LOGGER = logging.getLogger("ip_adapter")


def check_port(address, port):
    """
    Checks a given address and port to ensure that it is available. If not available, a ValueError is raised. Note: this
    is done by binding to an address. It does not call "listen"

    :param address: address that will bind to
    :param port: port to bind to
    """
    socket_trial = None
    try:
        socket_trial = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_trial.bind((address, port))
    except OSError as err:
        msg = f"Error with address/port of '{address}:{port}' : {err}"
        raise OSError(msg)
    finally:
        if socket_trial is not None:
            socket_trial.close()


class IpAdapter(fprime_gds.common.communication.adapters.base.BaseAdapter):
    """
    Adapts IP traffic for use with the GDS ground system. This serves two different "servers" both on the same address
    and port, but one uses TCP and the other uses UDP. Writes go to the TCP connection, and reads request data from
    both. This data is concatenated and returned up the stack for processing.
    """

    # Data to send out as part of the KEEPALIVE packet. Should not be null nor empty.
    KEEPALIVE_DATA = b"sitting well"
    MAXIMUM_DATA_SIZE = 4096

    def __init__(self, address, port, server=True, keepalive_interval=0.5):
        """
        Initialize this adapter by creating a handler for UDP and TCP. A thread for the KEEPALIVE application packets
        will be created, if the interval is not none. Handlers are servers unless server=False.
        """
        self.address = address
        self.port = port
        self.stop = False
        self.keepalive_thread = None
        self.keepalive_interval = keepalive_interval
        self.tcp = TcpHandler(address, port, server=server)
        self.udp = UdpHandler(address, port, server=server)
        self.thtcp = None
        self.thudp = None
        self.data_chunks = queue.Queue()
        self.blob = b""

    def __repr__(self):
        """ String representation for logging """
        return f"Paired-TCP/UDP@{self.address}:{self.port}"

    def open(self):
        """
        Open up the interface to the stored Address and Port. This will create a TCP and UDP socket. The TCP socket will
        be bound to the address, and listened to for incoming connects.
        """
        # Keep alive thread
        try:
            # Setup the tcp and udp adapter and run a thread to service them
            self.thtcp = threading.Thread(target=self.th_handler, name="TcpCommThread", args=(self.tcp,))
            self.thtcp.daemon = True
            self.thtcp.start()
            self.thudp = threading.Thread(target=self.th_handler, name="UdpCommThread", args=(self.udp,))
            self.thudp.daemon = True
            self.thudp.start()
            # Start up a keep-alive ping if desired. This will hit the TCP uplink, and die if the connection is down
            if self.keepalive_interval > 0.0:
                self.keepalive_thread = threading.Thread(
                    target=self.th_alive, name="KeepCommAliveThread", args=[self.keepalive_interval]
                )
                self.keepalive_thread.start()
        except (ValueError, TypeError) as exc:
            LOGGER.error(
                f"Failed to start keep-alive thread. {type(exc).__name__}: {str(exc)}"
            )

    def close(self):
        """Close the adapter, by setting the stop flag"""
        self.stop = True
        self.tcp.stop()
        self.udp.stop()

    def th_handler(self, handler):
        """Adapter thread function"""
        handler.open()
        while not self.stop:
            self.data_chunks.put(handler.read())
        handler.close()

    def write(self, frame):
        """
        Send a given framed bit of data by sending it out the serial interface. It will attempt to reconnect if there
        was a problem previously. This function will return true on success, or false on error.

        :param frame: framed data packet to send out
        :return: True, when data was sent through the UART. False otherwise.
        """
        if self.tcp.connected == IpHandler.CONNECTED:
            return self.tcp.write(frame)

    def read(self, timeout=0.500):
        """
        Read up to a given count in bytes from the TCP adapter. This may return less than the full requested size but
        is expected to return some data.

        :param timeout: timeout to wait for data. Needed as the get call below may not interrupt if it waits forever
        :return: data successfully read or "" when no data available within timeout
        """
        data = b""
        # The read function should block until data is available, but for efficiency, it should read all data available
        # thus the data is read, blocking for 0.5 seconds, and then reads as long as it is not empty.
        try:
            data += self.data_chunks.get(timeout=timeout)
            while not self.data_chunks.empty():
                data += self.data_chunks.get_nowait()
        except queue.Empty:
            pass
        return data

    def th_alive(self, interval):
        """
        Run this thread in order to accept incoming connections and spiral them off into another thread for handling the
        given client.
        """
        while not self.stop:
            self.write(IpAdapter.KEEPALIVE_DATA)
            time.sleep(interval)

    @classmethod
    def get_name(cls):
        """ Get the name of this adapter  """
        return "ip"

    @classmethod
    def get_arguments(cls):
        """
        Returns a dictionary of flag to argparse-argument dictionaries for use with argparse to setup arguments.

        :return: dictionary of flag to argparse arguments for use with argparse
        """
        return {
            ("--ip-address",): {
                "dest": "address",
                "type": str,
                "default": "0.0.0.0",
                "help": "Address of the IP adapter server.",
            },
            ("--ip-port",): {
                "dest": "port",
                "type": int,
                "default": 50000,
                "help": "Port of the IP adapter server.",
            },
            ("--ip-client",): {
                # dest is "server" since it is handled in BaseAdapter.construct_adapter and passed with the same
                # name to the IpAdapter constructor. Default to server=True, meaning IpAdapter is the TCP server
                "dest": "server",
                "action": "store_false",
                "default": True,
                "help": "Run the IP adapter as the client (connects to FSW running TcpServer)",
            },
            ("--keepalive-interval",): {
                "dest": "keepalive_interval",
                "type": float,
                "default": 0.5000,
                "help": "Keep alive packet interval. 0.0 = off, default = 0.5",
            },
        }

    @classmethod
    @gds_plugin_implementation
    def register_communication_plugin(cls):
        """ Register this as a communication plugin """
        return cls

    @classmethod
    def check_arguments(cls, address, port, server=True, keepalive_interval=0.5):
        """
        Code that should check arguments of this adapter. If there is a problem with this code, then a "ValueError"
        should be raised describing the problem with these arguments.

        :param args: arguments as dictionary
        """
        try:
            if server:
                check_port(address, port)
        except OSError as os_error:
            raise ValueError(f"{os_error}")



class IpHandler(abc.ABC):
    """
    Base handler for IP types. This will provide the basic methods, and synchronization for reading/writing to multiple
    child implementations, namely: UDP and TCP. These child objects can then be instantiated individually.
    """

    ERROR_RETRY_INTERVAL = (
        1  # Seconds between a non-timeout error and a socket reconnection
    )
    MAX_CLIENT_BACKLOG = 1  # One client backlog, allowing for reconnects
    # Connection states, it will go between these states
    CONNECTING = "CONNECTING"
    CONNECTED = "CONNECTED"
    CLOSED = "CLOSED"

    def __init__(
        self,
        address,
        port,
        adapter_type,
        server=True,
        logger=logging.getLogger("ip_handler"),
        post_connect=None,
    ):
        """
        Initialize this handler. This will set the variables, and start up the internal receive thread.

        :param address: address of the handler
        :param port: port of the handler
        :param adapter_type: type of this adapter. socket.SOCK_STREAM or socket.SOCK_DGRAM
        """
        self.running = True
        self.type = adapter_type
        self.address = address
        self.next_connect = 0
        self.port = port
        self.socket = None
        self.server = server
        self.connected = IpHandler.CLOSED
        self.logger = logger
        self.post_connect = post_connect
        atexit.register(self.stop)

    def open(self):
        """
        Open up this IP type adapter. Returning if already connected.
        """
        if self.CONNECTED == self.connected:
            return True
        # Just continually try and reconnect essentially "blocking" until open. Will only fail if we aborted.
        while self.running and self.CONNECTED != self.connected:
            try:
                # Prevent reconnects when the socket is connected. Socket should be closed on all errors
                if (
                    self.connected == IpHandler.CLOSED
                    and self.next_connect < time.time()
                ):
                    self.connected = IpHandler.CONNECTING
                    self.socket = socket.socket(socket.AF_INET, self.type)
                    if self.server:
                        self.socket.bind((self.address, self.port))
                    else:
                        self.socket.connect((self.address, self.port))
                    self.open_impl()
                    self.connected = IpHandler.CONNECTED
                    self.logger.info(
                        "%s connected to %s:%d",
                        "Server" if self.server else "Client",
                        self.address,
                        self.port,
                    )
                    # Post connect handshake
                    if self.post_connect is not None:
                        self.write(self.post_connect)
            # All errors (timeout included) we should close down the socket, which sets self.connected
            except ConnectionAbortedError:
                return False
            except OSError as exc:
                self.logger.warning(
                    "Failed to open socket at %s:%d, retrying: %s: %s",
                    self.address,
                    self.port,
                    type(exc).__name__,
                    str(exc),
                )
                self.next_connect = time.time() + IpHandler.ERROR_RETRY_INTERVAL
                self.close()
        return self.connected == self.CONNECTED

    @abc.abstractmethod
    def open_impl(self):
        """Implementation of the handler's open call"""

    def close(self):
        """
        Close this specific IP handler. This involves setting connected to False, and closing non-null sockets.
        """
        try:
            self.close_impl()
            IpHandler.kill_socket(self.socket)
        finally:
            self.socket = None
            self.connected = IpHandler.CLOSED

    @abc.abstractmethod
    def close_impl(self):
        """Implementation of the handler's close call"""

    def stop(self):
        """Stop the handler from reconnecting and close"""
        self.running = False
        self.close()

    def read(self):
        """
        Reads a single message after ensuring that the socket is fully open. On a non-timeout error, close the socket in
        preparation for a reconnect. This internally will call the child's read_impl

        :return: data read from TCP server or b"" when nothing is available
        """
        # This will block waiting for data
        try:
            return self.read_impl()
        except OSError as exc:
            if self.running:
                self.close()
                self.logger.warning(
                    "Read failure attempting reconnection. %s: %s",
                    type(exc).__name__,
                    str(exc),
                )
                self.open()
        return b""

    @abc.abstractmethod
    def read_impl(self):
        """Implementation of the handler's read call"""

    def write(self, message):
        """
        Writes a single message after ensuring that the socket is fully open. On any error, close the socket in
        preparation for a reconnect. This internally will call the child's write_impl

        :param message: message to send
        :return: True if all data was written, False otherwise
        """
        try:
            self.write_impl(message)
            return True
        except OSError as exc:
            if self.running:
                self.logger.warning(
                    "Write failure: %s: %s", type(exc).__name__, str(exc)
                )
        return False

    @abc.abstractmethod
    def write_impl(self, message):
        """Implementation of the handler's write call"""

    @staticmethod
    def kill_socket(sock):
        """Kills a socket connection, but shutting it down and then closing."""
        if sock is None:
            return
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        finally:
            sock.close()


class TcpHandler(IpHandler):
    """
    An IpAdapter that allows for interfacing with TCP socket.
    """

    def __init__(
        self,
        address,
        port,
        server=True,
        logger=logging.getLogger("tcp_handler"),
        post_connect=None,
    ):
        """
        Init the TCP adapter with port and address

        :param address: address of TCP
        :param port: port of TCP
        """
        super().__init__(
            address, port, socket.SOCK_STREAM, server, logger, post_connect
        )
        self.client = None
        self.client_address = None

    def open_impl(self):
        """
        Open up this particular adapter. This adapter
        """
        # When a server, must accept and spawn new socket
        if self.server:
            self.socket.listen(IpHandler.MAX_CLIENT_BACKLOG)
            (self.client, self.client_address) = self.socket.accept()
        # When a client, use normal socket
        else:
            self.client = self.socket

    def close_impl(self):
        """
        Close the TCP socket that was spawned as appropriate.
        """
        try:
            IpHandler.kill_socket(self.client)
        finally:
            self.client = None
            self.client_address = None

    def read_impl(self):
        """
        Specific read implementation for the TCP handler. This involves reading from the spawned client socket, not the
        primary socket.
        """
        data = self.client.recv(IpAdapter.MAXIMUM_DATA_SIZE)
        if not data:
            self.close_impl()
            self.open_impl()
        return data

    def write_impl(self, message):
        """
        Send is implemented with TCP. It will send it to the connected client.

        :param message: message to send out
        """
        # Block until the port is open
        while self.connected != IpHandler.CONNECTED or self.client is None:
            pass
        self.client.sendall(message)


class UdpHandler(IpHandler):
    """
    Handler for UDP traffic. This will work in unison with the TCP adapter.
    """

    def __init__(
        self, address, port, server=True, logger=logging.getLogger("udp_handler")
    ):
        """
        Init UDP with address and port

        :param address: address of UDP
        :param port: port of UDP
        """
        super().__init__(address, port, socket.SOCK_DGRAM, server, logger)

    def open_impl(self):
        """No extra steps required"""

    def close_impl(self):
        """No extra steps required"""

    def read_impl(self):
        """
        Receive from the UDP handler. This involves receiving from an unconnected socket.
        """
        (data, address) = self.socket.recvfrom(IpAdapter.MAXIMUM_DATA_SIZE)
        return data

    def write_impl(self, message):
        """
        Write not implemented with UDP
        """
        raise NotImplementedError("UDP Handler cannot send data.")



"""
@brief Base class for all loaders that load dictionaries from xml dictionaries

The XmlLoader class inherits from the DictLoader base class and is intended
to be inherited by loader classes that find its helper functions useful for
parsing xml dictionaries.

This class does not implement any core loader functions, it just adds common
helper functions

@date Created July 19, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

import os

from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType
from lxml import etree

from fprime_gds.common.utils.string_util import preprocess_c_style_format_str
from fprime_gds.common.data_types import exceptions
from fprime_gds.version import (
    MAXIMUM_SUPPORTED_FRAMEWORK_VERSION,
    MINIMUM_SUPPORTED_FRAMEWORK_VERSION,
)

# Custom Python Modules
from . import dict_loader


class XmlLoader(dict_loader.DictLoader):
    """Class to help load xml based dictionaries"""

    ENUM_SECT = "enums"
    ENUM_TYPE_TAG = "type"
    ENUM_SERIALIZE_TYPE_TAG = "serialize_type"
    ENUM_ELEM_NAME_TAG = "name"
    ENUM_ELEM_VAL_TAG = "value"
    ENUM_ELEM_DESC_TAG = "description"

    # Xml section names and tags for serializable types
    SER_SECT = "serializables"
    SER_TYPE_TAG = "type"
    SER_MEMB_SECT = "members"
    SER_MEMB_NAME_TAG = "name"
    SER_MEMB_FMT_STR_TAG = "format_specifier"
    SER_MEMB_DESC_TAG = "description"
    SER_MEMB_TYPE_TAG = "type"
    SER_MEMB_SIZE_TAG = "size"

    # Xml section names and tags for array types
    ARR_SECT = "arrays"
    ARR_NAME_TAG = "name"
    ARR_TYPE_TAG = "type"
    ARR_SIZE_TAG = "size"
    ARR_FORMAT_TAG = "format"
    ARR_DEFAULT_TAG = "defaults"
    ARR_DEFAULT_VALUE_TAG = "value"

    # Xml section names and tags for argument sections
    ARGS_SECT = "args"
    ARG_NAME_TAG = "name"
    ARG_DESC_TAG = "description"
    ARG_TYPE_TAG = "type"

    STR_LEN_TAG = "len"

    def __init__(self):
        """
        Constructor

        Returns:
            An initialized loader object
        """
        super().__init__()

        # These dicts hold already parsed enum objects so things don't need
        # to be parsed multiple times
        self.enums = {}
        self.serializable_types = {}
        self.array_types = {}

    @staticmethod
    def get_xml_tree(path):
        """
        Reads the xml file at the given path and parses it using lxml

        Args:
            path (string): Path to the xml dictionary file

        Returns:
            An lxml etree root object containing the parsed xml file
            information. Raises an exception if there is an error.
        """
        # Check that dictionary path exists
        if not os.path.isfile(path):
            raise exceptions.GseControllerUndefinedFileException(path)

        # Create xml parser
        xml_parser = etree.XMLParser(remove_comments=True)

        with open(path) as fd:
            # Parse xml and get element tree object we can retrieve data from
            element_tree = etree.parse(fd, parser=xml_parser)
        root = element_tree.getroot()
        return root

    @staticmethod
    def get_xml_section(section_name, xml_root):
        """
        Retrieve the given section in the xml tree if it exists

        Args:
            section_name (string): Section to retrieve
            xml_root (lxml etree root): xml object to search in

        Returns:
            The xml object of the desired section if found, or None if not
        """
        return next(
            (section for section in xml_root if section.tag == section_name), None
        )

    def get_args_list(self, xml_obj, xml_tree, context=None):
        """
        Parses and returns a standard xml dict arguments section:
          Section name: "args"
          Object tags: "arg"
          Object fields: "name", "type", "description"(optional)

        Args:
            xml_obj (lxml etree root): xml object containing the args section
                                       to parse.
            xml_tree (lxml etree root): Main xml tree object containing info on
                                        Enums and serializables.
            context: context for where we are in the tree

        Returns:
            List of arguments where each argument is a tuple of the form:
            (arg name [string], arg description [string or None], arg obj
            [python object derived from TypeBase]). If there is no args section
            or there are no arguments in the args section, [] is returned.
        """
        args = []
        args_section = self.get_xml_section(self.ARGS_SECT, xml_obj)

        if args_section is not None:
            for arg in args_section:
                arg_dict = arg.attrib

                arg_name = arg_dict[self.ARG_NAME_TAG]
                arg_type_name = arg_dict[self.ARG_TYPE_TAG]
                arg_typ_obj = self.parse_type(arg_type_name, arg, xml_tree, context)

                arg_desc = None
                if self.ARG_DESC_TAG in arg_dict:
                    arg_desc = arg_dict[self.ARG_DESC_TAG]

                args.append((arg_name, arg_desc, arg_typ_obj))

        return args

    def get_enum_type(self, enum_name, xml_obj):
        """
        Parses and returns an Enum object for the given enum name.

        Looks in the enums section of the xml dict.

        Args:
            enum_name (string): Name of the enum to parse
            xml_obj (lxml etree root): Parsed Xml object to find the enum in

        Returns:
            If the enum name could be found in the xml_obj, a corresponding
            object of type EnumType is returned. Otherwise, None is returned.
            The caller will hold the only reference to the object.
        """
        # Check if there is an already parsed version of this enum
        if enum_name in self.enums:
            return self.enums[enum_name]

        # Check if the dictionary has an enum section
        enum_section = self.get_xml_section(self.ENUM_SECT, xml_obj)
        if enum_section is None:
            return None

        for enum in enum_section:
            # Check enum name
            if enum.get(self.ENUM_TYPE_TAG) == enum_name:
                # Get serialize/representation type, if present
                serialize_type = enum.get(self.ENUM_SERIALIZE_TYPE_TAG, "I32")

                # Go through all possible values of the enum
                members = {}
                for item in enum:
                    item_name = item.get(self.ENUM_ELEM_NAME_TAG)
                    item_val = int(item.get(self.ENUM_ELEM_VAL_TAG))
                    members[item_name] = item_val

                enum_obj = EnumType.construct_type(enum_name, members, serialize_type)

                self.enums[enum_name] = enum_obj
                return enum_obj

        return None

    def get_serializable_type(self, type_name, xml_obj):
        """
        Parses and returns a serializable type object for the given type name.

        Looks in the serializables section of the xml dict.

        Args:
            type_name (string): Name of the type to parse
            xml_obj (lxml etree root): Parsed Xml object to find type in

        Returns:
            If the type name could be found in the xml_obj, a corresponding
            object of a type derived from SerializableType is returned.
            Otherwise, None is returned. The caller will hold the only reference
            to the object.
        """
        # Check if there is already a parsed version of this serializable
        if type_name in self.serializable_types:
            return self.serializable_types[type_name]

        # Check if the dictionary has an enum section
        ser_section = self.get_xml_section(self.SER_SECT, xml_obj)
        if ser_section is None:
            return None

        for ser_type in ser_section:
            # Check if this serializable matches the type name
            if ser_type.get(self.SER_TYPE_TAG) == type_name:
                # Go through members
                memb_section = self.get_xml_section(self.SER_MEMB_SECT, ser_type)

                # If there is no member section, this type is invalid
                if memb_section is None:
                    return None

                members = []
                for memb in memb_section:
                    name = memb.get(self.SER_MEMB_NAME_TAG)
                    fmt_str = XmlLoader.preprocess_format_str(
                        memb.get(self.SER_MEMB_FMT_STR_TAG)
                    )
                    desc = memb.get(self.SER_MEMB_DESC_TAG)
                    memb_type_name = memb.get(self.SER_MEMB_TYPE_TAG)
                    memb_size = memb.get(self.SER_MEMB_SIZE_TAG)
                    type_obj = self.parse_type(memb_type_name, memb, xml_obj)
                    # memb_size is not None for member array
                    if memb_size is not None:
                        type_obj = ArrayType.construct_type(
                            f"Array_{type_obj.__name__}_{memb_size}",
                            type_obj,
                            int(memb_size),
                            fmt_str,
                        )

                    members.append((name, type_obj, fmt_str, desc))

                ser_obj = SerializableType.construct_type(type_name, members)

                self.serializable_types[type_name] = ser_obj
                return ser_obj

        return None

    def get_array_type(self, type_name, xml_obj):
        """
        Parses and returns an array type object for the given type name.

        Looks in the arrays section of the xml dict.

        Args:
            type_name (string): Name of the type to parse
            xml_obj (lxml etree root): Parsed Xml object to find type in

        Returns:
            If the type name could be found in the xml_obj, a corresponding
            object of a type derived from ArrayType is returned.
            Otherwise, None is returned. The caller will hold the only reference
            to the object.
        """

        # Check if there is already a parsed version of this array
        if type_name in self.array_types:
            return self.array_types[type_name]

        # Check if the dictionary has an array section
        arr_section = self.get_xml_section(self.ARR_SECT, xml_obj)
        if arr_section is None:
            return None

        for arr_memb in arr_section:
            # Check if this array matches the name name
            if arr_memb.get(self.ARR_NAME_TAG) == type_name:
                # Go through default members
                default_section = self.get_xml_section(self.ARR_DEFAULT_TAG, arr_memb)

                # If there is no default member section, this type is invalid
                if default_section is None:
                    return None

                # Make config
                arr_type = arr_memb.get(self.ARR_TYPE_TAG)
                type_obj = self.parse_type(arr_type, arr_memb, xml_obj)
                arr_format = XmlLoader.preprocess_format_str(
                    arr_memb.get(self.ARR_FORMAT_TAG)
                )
                arr_size = arr_memb.get(self.ARR_SIZE_TAG)

                arr_obj = ArrayType.construct_type(
                    type_name, type_obj, int(arr_size), arr_format
                )

                self.array_types[type_name] = arr_obj
                return arr_obj

        return None

    def parse_type(self, type_name, xml_item, xml_tree, context=None):
        """
        Parses the given type string and returns a type object.

        Args:
            type_name (string): Name of the type in the xml
            xml_item (lxml etree root): Parsed xml object for the item
                      containing the type name being parsed. This is used to get
                      meta data such as string length.
            xml_tree (lxml etree root): Parsed Xml object containing enum and
                                        serializable type info (may not be used)
            context: context in the tree

        Returns:
            Object of a class derived from the TypeBase class if successful,
            Raises an exception if the parsing fails. The caller will hold the
            only reference to the object.
        """

        if type_name == "I8":
            return I8Type
        if type_name == "I16":
            return I16Type
        if type_name == "I32":
            return I32Type
        if type_name == "I64":
            return I64Type
        if type_name == "U8":
            return U8Type
        if type_name == "U16":
            return U16Type
        if type_name == "U32":
            return U32Type
        if type_name == "U64":
            return U64Type
        if type_name == "F32":
            return F32Type
        if type_name == "F64":
            return F64Type
        if type_name == "bool":
            return BoolType
        if type_name == "string":
            if self.STR_LEN_TAG not in xml_item.attrib:
                print(
                    f"Trying to parse string type, but found {self.STR_LEN_TAG} field"
                )
                return None
            max_length = int(xml_item.get(self.STR_LEN_TAG, 0))
            name = f"{context or ''}::{xml_item.get(self.ARG_NAME_TAG)}String"
            return StringType.construct_type(name, max_length)
        # First try Serialized types:
        result = self.get_serializable_type(type_name, xml_tree)
        if result is not None:
            return result

        # Now try enums:
        result = self.get_enum_type(type_name, xml_tree)
        if result is not None:
            return result

        # Now try arrays:
        result = self.get_array_type(type_name, xml_tree)
        if result is not None:
            return result

        # Abandon all hope
        msg = f"Could not find type {type_name}"
        raise exceptions.GseControllerParsingException(msg)

    @staticmethod
    def preprocess_format_str(format_str):
        """Converts C-style format strings to Python-style format strings
        For example "%x" -> "{:x}" or "%.2f" -> "{:.2f}"

        Args:
            format_str (str): C-style format string

        Returns:
            str: Python-style format string
        """
        return preprocess_c_style_format_str(format_str)



"""
@brief Decoder for event data

This decoder takes in serialized events, parses them, and packages the results
in event_data objects.

Example data structure:
    +-------------------+---------------------+---------------------- - - -
    | ID (4 bytes)      | Time Tag (11 bytes) | Event argument data....
    +-------------------+---------------------+---------------------- - - -

@date Created June 29, 2018
@author R. Joseph Paetz

@bug No known bugs
"""
from fprime.common.models.serialize import time_type
from fprime.common.models.serialize.type_exceptions import TypeException

from fprime_gds.common.data_types import event_data
from fprime_gds.common.decoders import decoder
from fprime_gds.common.decoders.decoder import DecodingException
from fprime_gds.common.utils import config_manager

import logging
LOGGER = logging.getLogger("event_decoder")

class EventDecoder(decoder.Decoder):
    """Decoder class for event data"""

    def __init__(self, event_dict, config=None):
        """
        EventDecoder class constructor

        Args:
            event_dict: Event dictionary. Event IDs should be keys and
                        EventTemplate objects should be values

        Returns:
            An initialized EventDecoder object.
        """
        super(EventDecoder, self).__init__()

        if config is None:
            # Retrieve defaults for the configs
            config = config_manager.ConfigManager.get_instance()

        self.__dict = event_dict
        self.id_obj = config.get_type("FwEventIdType")

    def decode_api(self, data):
        """
        Decodes the given data and returns the result.

        This function allows for non-registered code to call the same decoding
        code as is used to parse data passed to the data_callback function.

        Args:
            data: Binary data to decode

        Returns:
            Parsed version of the event data in the form of a EventData object
            or None if the data is not decodable
        """
        ptr = 0

        event_list = []

        while (ptr < len(data)):

            # Decode event ID here...
            if ptr + self.id_obj.getSize() <= len(data):
                self.id_obj.deserialize(data, ptr)
            else:
                LOGGER.warning("Insufficient data for event ID")
                break
            ptr += self.id_obj.getSize()
            event_id = self.id_obj.val

            # Decode time...
            event_time = time_type.TimeType()
            event_time.deserialize(data, ptr)
            ptr += event_time.getSize()

            if event_id not in self.__dict:
                msg = f"Event {event_id} not found in dictionary"
                raise DecodingException(msg)

            event_temp = self.__dict[event_id]

            (size, arg_vals) = self.decode_args(data, ptr, event_temp)

            event_list.append(event_data.EventData(arg_vals, event_time, event_temp))
            # add up argument sizes
            ptr += size
        return event_list

    @staticmethod
    def decode_args(arg_data, offset, template):
        """
        Decodes the serialized event arguments

        The event arguments are each serialized and then appended to each other.
        Parse that section of the data into the individual arguments.

        Args:
            arg_data: Serialized argument data to parse
            offset: Offset into the arg_data where parsing should start
            template: EventTemplate object that describes the type of event the
                      arg_data goes to.

        Returns:
            Parsed arguments in a tuple (order the same as they were parsed in).
            Each element in the tuple is an instance of the same class as the
            corresponding arg_type object in the template parameter. Returns
            none if the arguments can't be parsed
        """
        arg_results = []
        args = template.get_args()

        # For each argument, use the arg_obj deserialize method to get the value
        for arg in args:
            (arg_name, arg_desc, arg_type) = arg

            arg_obj = arg_type()

            try:
                arg_obj.deserialize(arg_data, offset)
                arg_results.append(arg_obj)
            except TypeException as e:
                msg = f"Event argument decoding failed {e.getMsg()}"
                raise DecodingException(msg)

            offset = offset + arg_obj.getSize()

        return [offset, tuple(arg_results)]



import math
import pytest
from fprime_gds.common.fpy.codegen import (
    SPECIFIC_INTEGER_TYPES,
    SPECIFIC_NUMERIC_TYPES,
    UNSIGNED_INTEGER_TYPES,
    SIGNED_INTEGER_TYPES,
    SPECIFIC_FLOAT_TYPES,
    FppTypeClass,
)
from fprime.common.models.serialize.numerical_types import I64Type, U64Type, F64Type

from fprime_gds.common.fpy.model import MIN_INT64, overflow_check
from fprime_gds.common.fpy.test_helpers import (
    assert_compile_failure,
    assert_run_success,
)

MAX_BITWIDTH_TYPES = [I64Type, U64Type, F64Type]
NUMERIC_VALUES = ["min", "-1", "0", "1", "max"]

ARITHMETIC_OPERATORS = [
    "+",
    "-",
    "/",
    "*",
]


def get_max(type: FppTypeClass) -> int | float:
    assert type in SPECIFIC_NUMERIC_TYPES
    if type in SPECIFIC_INTEGER_TYPES:
        return type.range()[1] - 1

    # otherwise, return float or double max

    if type.get_bits() == 32:
        # f32
        return 3.402823e38
    # f64
    return 1.79769e308


def get_min(type: FppTypeClass) -> int | float:
    assert type in SPECIFIC_NUMERIC_TYPES
    if type in SPECIFIC_INTEGER_TYPES:
        return type.range()[0]

    # otherwise, return float or double min

    if type.get_bits() == 32:
        # f32
        return -3.402823e38
    # f64
    return -1.79769e308


def get_fpy_str(type: FppTypeClass) -> str:
    assert type in SPECIFIC_NUMERIC_TYPES
    return type.get_canonical_name()


def get_val(type: FppTypeClass, val_str: str) -> int | float:
    assert type in SPECIFIC_NUMERIC_TYPES
    if val_str == "max":
        return get_max(type)
    if val_str == "min":
        return get_min(type)

    if type in SPECIFIC_INTEGER_TYPES:
        return int(val_str)
    return float(val_str)


@pytest.mark.parametrize("lhs_type", MAX_BITWIDTH_TYPES)
@pytest.mark.parametrize("rhs_type", MAX_BITWIDTH_TYPES)
@pytest.mark.parametrize("lhs_val", NUMERIC_VALUES)
@pytest.mark.parametrize("rhs_val", NUMERIC_VALUES)
@pytest.mark.parametrize("op", ARITHMETIC_OPERATORS)
def test_math_with_max_bitwidth_types(
    fprime_test_api, lhs_type, rhs_type, lhs_val, rhs_val, op
):
    lhs_val = get_val(lhs_type, lhs_val)
    rhs_val = get_val(rhs_type, rhs_val)
    seq = ""
    # set lhs and rhs vars
    seq += "lhs: " + lhs_type.get_canonical_name() + " = " + str(lhs_val) + "\n"
    seq += "rhs: " + rhs_type.get_canonical_name() + " = " + str(rhs_val) + "\n"

    if not (get_min(lhs_type) <= lhs_val <= get_max(lhs_type)) or not (
        get_min(rhs_type) <= rhs_val <= get_max(rhs_type)
    ):
        print(lhs_type, get_min(lhs_type))
        # not representable. seq should fail compile
        assert_compile_failure(fprime_test_api, seq)
        return

    result_type = None
    if F64Type in (lhs_type, rhs_type):
        result_type = F64Type
    elif U64Type in (lhs_type, rhs_type):
        result_type = U64Type
    else:
        result_type = I64Type

    should_fail = False

    ans = 0
    if op == "/":
        if rhs_val == 0:
            should_fail = True
            ans = math.inf
        else:
            if result_type == F64Type:
                ans = lhs_val / rhs_val
            else:
                if lhs_val == MIN_INT64 and rhs_val == -1:
                    # cpp specific overflow behavior
                    ans = MIN_INT64
                else:
                    # do int division
                    ans = int(lhs_val / rhs_val)
    elif op == "+":
        ans = lhs_val + rhs_val
    elif op == "-":
        ans = lhs_val - rhs_val
    elif op == "*":
        ans = lhs_val * rhs_val
    else:
        assert False, op

    if result_type != F64Type and op != "/":
        # if integer and not division, prevent overflow
        ans = overflow_check(ans)

    # in a perfect world, this is the answer
    # but this is not a perfect world.

    if math.isinf(ans):
        should_fail = True

    seq += "if lhs " + op + " rhs == " + str(ans) + ":\n"
    seq += "    exit(True)\n"
    seq += "exit(False)\n"
    print(seq)
    if should_fail:
        assert_compile_failure(fprime_test_api, seq)
    else:
        assert_run_success(fprime_test_api, seq)


import re
from datetime import datetime, timedelta

from fprime_gds.common.data_types import exceptions as gseExceptions
from fprime_gds.common.models.common.command import Descriptor


class SeqFileParser:
    def parse(self, filename, cont=False):
        """
        Generator that parses an input sequence file and returns a tuple
        for each valid line of the sequence file.
        @param seqfile: A sequence file name (usually a .seq extension)
        @param cont: attempt to continue after a line fails to parse, hopefully revealing more errors
        @return A list of tuples:
            (lineNumber, descriptor, seconds, useconds, mnemonic, arguments)
        """

        def subQuoted(f, string):
            """
            Run a substitution function on only substrings within a string that are surrounded
            by single or double quotes
            @param f: a string substitution acting on matchobjs
            @param string: the string to perform the substitution on
            @return the substituted string
            """
            s = re.sub(r'"[^"]*"', f, string)
            return re.sub(r"'[^']*'", f, s)

        def removeTrailingComments(string):
            """
            Remove any trailing comments (preceded by ';') in a string
            @param string: the string to perform comment removal on
            @return the string without trailing comments
            """

            def replaceSemis(matchobj):
                return matchobj.group(0).replace(";", " ")

            # ignore all semicolons in quotes:
            s = subQuoted(replaceSemis, string)
            s = subQuoted(replaceSemis, s)
            # get index of first semicolon, and return everything before it:
            if ";" in s:
                index = s.index(";")
                return string[:index]
            # return original string if no semicolon found:
            return string

        def splitString(string):
            """
            Split a string with ' ' or ',' as a delimiter. Ignore any delimiters
            found within quoted substrings.
            @param string: the string to perform the split on
            @return a list representing the split string
            """

            def replaceSpacesAndCommas(matchobj):
                s = re.sub(r"\s", "_", matchobj.group(0))
                s = re.sub(r"\,", "_", s)
                return s

            # ignore all spaces in quotes:
            s = subQuoted(replaceSpacesAndCommas, string)
            # replace all commas with spaces, since either can be a delimiter:
            s = s.replace(",", " ")
            # get the split indices of the modified string:
            indices = [(m.start(), m.end()) for m in re.finditer(r"\S+", s)]
            return [string[start:end] for start, end in indices]

        def parseArgs(args):
            """
            Turn .seq command argument list into their appropriate python types
            @param args: a list of parsed arguments
            @return a list of arguments as native python types
            """

            def parseArg(arg):
                # See if argument is a string, if so remove the quotes and return it:
                if (arg[0] == '"' and arg[-1] == '"') or (
                    arg[0] == "'" and arg[-1] == "'"
                ):
                    return arg[1:-1]
                # If the string contains a "." assume that it is a float:
                if "." in arg:
                    return float(arg)
                if arg in ["True", "true", "TRUE"]:
                    return True
                if arg in ["False", "false", "FALSE"]:
                    return False
                try:
                    # See if it translates to an integer:
                    return int(arg, 0)
                except ValueError:
                    try:
                        # See if it translates to a float:
                        return float(arg)
                    except ValueError:
                        # Otherwise it is an enum type:
                        return str(arg)

            return list(map(parseArg, args))

        def parseTime(lineNumber, time):
            """
            Parse a time string and return the command descriptor, seconds, and useconds of the time string
            @param lineNumber: the current line number where the time string was parsed
            @param time: the time string to parse
            @return a tuple (descriptor, seconds, useconds)
            """

            def parseTimeStringOption(timeStr, timeFmts):
                """
                Parse a time string by trying to use different time formats, until one succeeds
                @param timeStr: the time string
                @param timeFmts: the time format used to parse the string
                @return the datetime object containing the parsed string
                """

                def parseTimeString(timeFmt):
                    try:
                        return datetime.strptime(timeStr, timeFmt)
                    except ValueError:
                        return None

                for fmt in timeFmts:
                    dt = parseTimeString(fmt)
                    if dt:
                        return dt
                raise Exception("Failed to parse time string")

            def parseRelative(timeStr):
                """
                Parse a relative time string
                @param timeStr: the time string
                @return the datetime object containing the parsed string
                """
                options = ["%H:%M:%S.%f", "%H:%M:%S"]
                return parseTimeStringOption(timeStr, options)

            def parseAbsolute(timeStr):
                """
                Parse an absolute time string
                @param timeStr: the time string
                @return the datetime object containing the parsed string
                """
                options = ["%Y-%jT%H:%M:%S.%f", "%Y-%jT%H:%M:%S"]
                return parseTimeStringOption(timeStr, options)

            descriptor = None
            d = time[0]
            t = time[1:]
            if d == "R":
                descriptor = Descriptor.RELATIVE
                dt = parseRelative(t)
                delta = timedelta(
                    hours=dt.hour,
                    minutes=dt.minute,
                    seconds=dt.second,
                    microseconds=dt.microsecond,
                ).total_seconds()
            elif d == "A":
                descriptor = Descriptor.ABSOLUTE
                dt = parseAbsolute(t)
                # See if timezone was specified. If not, use UTC
                if dt.tzinfo is not None:
                    print(f"Using timezone {dt.tzinfo.tzname()}")
                    epoch = datetime.fromtimestamp(0, dt.tzinfo)
                else:
                    print("Using UTC timezone")
                    epoch = datetime.utcfromtimestamp(0)
                delta = (dt - epoch).total_seconds()
            else:
                msg = f"Line {lineNumber + 1}: Invalid time descriptor {d} found. Descriptor should either be 'A' for absolute times or 'R' for relative times"
                raise gseExceptions.GseControllerParsingException(
                    msg
                )
            seconds = int(delta)
            useconds = int((delta - seconds) * 1000000)
            return descriptor, seconds, useconds

        # Open the sequence file and parse each line:
        with open(filename) as inputFile:
            messages = []
            for i, line in enumerate(inputFile):
                try:
                    line = line.strip()
                    # ignore blank lines and comments
                    if line and line[0] != ";":
                        line = removeTrailingComments(line)
                        line = splitString(line)
                        length = len(line)
                        if length < 2:
                            msg = f'Line {i + 1}: Each line must contain a minimum of two fields, time and command mnemonic\n'
                            raise gseExceptions.GseControllerParsingException(
                                msg
                            )
                        try:
                            descriptor, seconds, useconds = parseTime(i, line[0])
                        except Exception:
                            msg = f'Line {i + 1}: Encountered syntax error parsing timestamp'
                            raise gseExceptions.GseControllerParsingException(
                                msg
                            )
                        mnemonic = line[1]
                        args = []
                        if length > 2:
                            args = line[2:]
                            try:
                                args = parseArgs(args)
                            except Exception:
                                msg = f'Line {i + 1}: Encountered syntax error parsing arguments'
                                raise gseExceptions.GseControllerParsingException(
                                    msg
                                )
                        yield i, descriptor, seconds, useconds, mnemonic, args
                except gseExceptions.GseControllerParsingException as exc:
                    if not cont:
                        raise
                    messages.append(exc.getMsg())
            if cont and messages:
                raise gseExceptions.GseControllerParsingException("\n".join(messages))



"""
pkt_json_loader.py:

Loads flight dictionary (JSON) and returns Python dictionaries of telemetry packets

@author jawest
"""

from fprime_gds.common.templates.pkt_template import PktTemplate
from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


class PktJsonLoader(JsonLoader):
    """Class to load python based telemetry packet dictionaries"""

    PACKETS_FIELD = "telemetryPacketSets"

    SET_NAME = "name"
    MEMBERS = "members"

    def get_packet_set_names(self, path):
        """ Get the list of packet sets """
        return [packet_set[self.SET_NAME] for packet_set in self.json_dict[self.PACKETS_FIELD]]

    def get_id_dict(self, path, packet_set_name: str, ch_name_dict: dict):
        if path in self.saved_dicts and packet_set_name in self.saved_dicts[path]:
            (id_dict, name_dict) = self.saved_dicts[path][packet_set_name]
        else:
            (id_dict, name_dict, self.versions) = self.construct_dicts(packet_set_name, ch_name_dict)
            if path not in self.saved_dicts:
                self.saved_dicts[path] = dict()
            self.saved_dicts[path].update({packet_set_name: (id_dict, name_dict)})

        return id_dict
    
    def get_name_dict(self, path, packet_set_name: str, ch_name_dict: dict):
        if path in self.saved_dicts and packet_set_name in self.saved_dicts[path]:
            (id_dict, name_dict) = self.saved_dicts[path][packet_set_name]
        else:
            (id_dict, name_dict, self.versions) = self.construct_dicts(packet_set_name, ch_name_dict)
            if path not in self.saved_dicts:
                self.saved_dicts[path] = dict()
            self.saved_dicts[path].update({packet_set_name: (id_dict, name_dict)})

        return name_dict


    def construct_dicts(self, packet_set_name: str, ch_name_dict: dict):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            ch_name_dict (dict()): Channel dictionary with names as keys and
                                   ChTemplate objects as values.

        Returns:
            A tuple with two packet dictionaries (type==dict()):
            (id_dict, name_dict) and the dictionary version. The keys of the packet dictionaries should 
            be the packets' id and name fields respectively and the values should be PktTemplate objects.
        """
        id_dict = {}
        name_dict = {}

        if self.PACKETS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.PACKETS_FIELD}' field: {str(self.json_file)}"
            )

        for packet_dict in self.json_dict[self.PACKETS_FIELD]:
            try:
                if packet_set_name == packet_dict[self.SET_NAME]:
                    for packet_group_dict in packet_dict.get(self.MEMBERS, []):
                        packet_temp = self.construct_template_from_dict(packet_group_dict, ch_name_dict)
                        id_dict[packet_temp.get_id()] = packet_temp
                        name_dict[packet_temp.get_name()] = packet_temp

                    return (
                        dict(sorted(id_dict.items())),
                        dict(sorted(name_dict.items())),
                        self.get_versions(),
                    )
                
            except KeyError as e:
                raise GdsDictionaryParsingException(
                    f"{str(e)} key missing from telemetry packet dictionary entry: {str(packet_dict)}"
                )
            
        raise GdsDictionaryParsingException(
            f"Ground Dictionary does not contain packet set '{packet_set_name}'"
        )

    def construct_template_from_dict(self, packet_group_dict: dict, ch_name_dict: dict):
        """        
        Args:
            packet_group_dict (dict()): Packet group dictionary with group id, name, and members
            ch_name_dict (dict()): Channel dictionary with names as keys and ChTemplate objects as values.
        Returns:
            A a PktTemplate object containing the packet group id, group name, and list of ChTemplate 
            objects that represent each member in the packet.
        """
        try:
            ch_list = []
            group_name = packet_group_dict["name"]
            group_id = packet_group_dict["id"]
            group_members = packet_group_dict["members"]

            for ch_name in group_members:
                ch_template = ch_name_dict[ch_name]
                ch_list.append(ch_template)
            
        except KeyError as e:
            raise GdsDictionaryParsingException(
                f"{str(e)} key missing from telemetry packet member or member is not a channel in the dictionary: {str(group_name)}"
            )
        
        return PktTemplate(
            group_id,
            group_name,
            ch_list
        )




# compiler debug flag
from dataclasses import dataclass
import traceback
from typing import Any

from lark import LarkError, Token, UnexpectedToken
from lark.indenter import DedentError

# assigned in compiler_main
file_name = None
# assigned in compiler_main
debug = False
# assigned in parse 
input_text = None
# assigned in parse 
input_lines = None


# the number of lines to show around a compiler error
COMPILER_ERROR_CONTEXT_LINE_COUNT = 1


@dataclass
class CompileError:
    msg: str
    node: Any = None

    def __post_init__(self):
        self.stack_trace = "\n".join(traceback.format_stack(limit=8)[:-1])

    def __repr__(self):

        stack_trace_optional = f"{self.stack_trace}\n" if debug else ""
        file_name_optional = (
            f"{file_name}" if file_name is not None else "<unknown file>"
        )

        if self.node is None:
            return f"{stack_trace_optional}{file_name_optional}: {self.msg}"

        meta = self.node if isinstance(self.node, Token) else self.node.meta

        source_start_line = meta.line - 1 - COMPILER_ERROR_CONTEXT_LINE_COUNT
        source_start_line = max(0, source_start_line)
        source_end_line = (
            meta.end_line - 1 + COMPILER_ERROR_CONTEXT_LINE_COUNT
        )
        source_end_line = min(len(input_lines), source_end_line)

        # this is the list of all the src lines we will display
        source_to_display: list[str] = input_lines[
            source_start_line : source_end_line + 1
        ]

        # reserve this much space for the line numbers
        line_number_space = 4 if source_end_line < 998 else 8

        # prefix all the lines with the prefix and line number
        # right justified line number, then a |, then the line
        source_to_display = [
            str(source_start_line + line_idx + 1).rjust(line_number_space)
            + " | "
            + line
            for line_idx, line in enumerate(source_to_display)
        ]

        node_lines = meta.end_line - meta.line

        if node_lines > 1:
            source_to_display_str = "\n".join(source_to_display)
            # it's a multiline node. don't try to highlight the whole thing
            # just print the err and the offending text
            return f"{stack_trace_optional}{file_name_optional}:{meta.line}-{meta.end_line} {self.msg}\n{source_to_display_str}"

        node_start_line_in_ctx = meta.line - 1 - source_start_line
        error_highlight = " " * (
            meta.column - 1 + line_number_space + 3
        ) + "^" * (meta.end_column - meta.column)
        source_to_display.insert(node_start_line_in_ctx + 1, error_highlight)
        result = f"{stack_trace_optional}{file_name_optional}:{meta.line} {self.msg}\n"
        result += "\n".join(source_to_display)

        return result


def handle_lark_error(err):
    assert isinstance(err, LarkError), err
    if isinstance(err, UnexpectedToken):
        print(str(CompileError("Invalid syntax", err.token)))
    elif isinstance(err, DedentError):
        print(str(CompileError(err.args[0])))
    exit(1)



"""
ch_json_loader.py:

Loads flight dictionary (JSON) and returns id and mnemonic based Python dictionaries of channels

@author thomas-bc
"""

from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.loaders.json_loader import JsonLoader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


class ChJsonLoader(JsonLoader):
    """Class to load python based telemetry channel dictionaries"""

    CHANNELS_FIELD = "telemetryChannels"

    ID = "id"
    NAME = "name"
    DESC = "annotation"
    TYPE = "type"
    FMT_STR = "format"
    LIMIT_FIELD = "limit"
    LIMIT_LOW = "low"
    LIMIT_HIGH = "high"
    LIMIT_RED = "red"
    LIMIT_ORANGE = "orange"
    LIMIT_YELLOW = "yellow"

    def construct_dicts(self, _):
        """
        Constructs and returns python dictionaries keyed on id and name

        Args:
            _: Unused argument (inherited)
        Returns:
            A tuple with two channel dictionaries (python type dict):
            (id_dict, name_dict). The keys should be the channels' id and
            name fields respectively and the values should be ChTemplate
            objects.
        """
        id_dict = {}
        name_dict = {}

        if self.CHANNELS_FIELD not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Ground Dictionary missing '{self.CHANNELS_FIELD}' field: {str(self.json_file)}"
            )

        for ch_dict in self.json_dict[self.CHANNELS_FIELD]:
            # Create a channel template object
            ch_temp = self.construct_template_from_dict(ch_dict)

            id_dict[ch_temp.get_id()] = ch_temp
            name_dict[ch_temp.get_full_name()] = ch_temp

        return (
            dict(sorted(id_dict.items())),
            dict(sorted(name_dict.items())),
            self.get_versions(),
        )

    def construct_template_from_dict(self, channel_dict: dict) -> ChTemplate:
        try:
            ch_id = channel_dict[self.ID]
            # The below assignment also raises a ValueError if the name does not contain a '.'
            qualified_component_name, channel_name = channel_dict[self.NAME].rsplit('.', 1)
            if not qualified_component_name or not channel_name:
                raise ValueError()

            type_obj = self.parse_type(channel_dict[self.TYPE])
        except ValueError as e:
            raise GdsDictionaryParsingException(
                f"Channel dictionary entry malformed, expected name of the form '<QUAL_COMP_NAME>.<CH_NAME>' in : {str(channel_dict)}"
            )
        except KeyError as e:
            raise GdsDictionaryParsingException(
                f"{str(e)} key missing from Channel dictionary entry or its associated type in the dictionary: {str(channel_dict)}"
            )

        format_str = JsonLoader.preprocess_format_str(channel_dict.get(self.FMT_STR))

        limit_field = channel_dict.get(self.LIMIT_FIELD)
        limit_low = limit_field.get(self.LIMIT_LOW) if limit_field else None
        limit_high = limit_field.get(self.LIMIT_HIGH) if limit_field else None
        limit_low_yellow = limit_low.get(self.LIMIT_YELLOW) if limit_low else None
        limit_low_red = limit_low.get(self.LIMIT_RED) if limit_low else None
        limit_low_orange = limit_low.get(self.LIMIT_ORANGE) if limit_low else None
        limit_high_yellow = limit_high.get(self.LIMIT_YELLOW) if limit_high else None
        limit_high_red = limit_high.get(self.LIMIT_RED) if limit_high else None
        limit_high_orange = limit_high.get(self.LIMIT_ORANGE) if limit_high else None

        return ChTemplate(
            ch_id,
            channel_name,
            qualified_component_name,
            type_obj,
            ch_fmt_str=format_str,
            ch_desc=channel_dict.get(self.DESC),
            low_red=limit_low_red,
            low_orange=limit_low_orange,
            low_yellow=limit_low_yellow,
            high_yellow=limit_high_yellow,
            high_orange=limit_high_orange,
            high_red=limit_high_red,
        )



"""ccsds.apid: APID mapping functions for F´ data"""

from fprime_gds.common.utils.data_desc_type import DataDescType
from fprime.common.models.serialize.numerical_types import NumericalType


class APID(object):
    """APID implementations"""

    # TODO: use the DataDescType configured by loading the dictionary

    @classmethod
    def from_type(cls, data_type: DataDescType):
        """Map from data description type to APID"""
        return data_type.value

    @classmethod
    def from_data(cls, data, packet_descriptor_type: NumericalType):
        """Map from data bytes to APID"""
        packet_descriptor_type.deserialize(data, offset=0)
        return cls.from_type(DataDescType(packet_descriptor_type.val))



"""
Tests Command Data

Created on Jan 11, 2025
@author: AlesKus
"""

import unittest
from fprime.common.models.serialize.bool_type import BoolType
import fprime.common.models.serialize.numerical_types as numerical_types
from fprime.common.models.serialize.string_type import StringType
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.data_types.cmd_data import CmdData


class CmdDataTest(unittest.TestCase):
    def test_str_bool_command_arg(self):
        """
        Tests CmdData::get_str() for a command with one bool argument
        """
        opcode = 42
        mnemonic = "TEST_CMD_BOOL"
        component = "PythonTests"
        arguments = [("status", None, BoolType)]
        command_template = CmdTemplate(opcode, mnemonic, component, arguments)
        cmd_data = CmdData((True, ), command_template)

        data_str = cmd_data.get_str()

        self.assertIn("[True]", data_str)

    def test_str_numerical_command_args(self):
        """
        Tests CmdData::get_str() for a command with several numerical arguments
        """
        opcode = 43
        mnemonic = "TEST_CMD_NUMERICS"
        component = "PythonTests"
        arguments = [("just_int", None, numerical_types.U32Type),
                     ("just_float", None, numerical_types.F64Type),
                     ("just_byte", None, numerical_types.I8Type)]
        command_template = CmdTemplate(opcode, mnemonic, component, arguments)
        cmd_data = CmdData((12345, 98.765, 127), command_template)

        data_str = cmd_data.get_str()

        self.assertIn("[12345, 98.765, 127]", data_str)

    def test_str_string_command_args(self):
        """
        Tests CmdData::get_str() for a command with one string argument
        """
        opcode = 44
        mnemonic = "TEST_CMD_STR"
        component = "PythonTests"
        arguments = [("just_string", None, StringType.construct_type(f"String_{128}", 128))]
        command_template = CmdTemplate(opcode, mnemonic, component, arguments)
        cmd_data = CmdData(("Everything is good", ), command_template)

        data_str = cmd_data.get_str()

        self.assertIn("['Everything is good']", data_str)


if __name__ == "__main__":
    unittest.main()



"""
json_loader.py:

Base class for all loaders that load dictionaries from json dictionaries

@author thomas-bc
"""

import json
from typing import Optional

from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
import fprime.common.models.serialize.numerical_types as numerical_types
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.type_base import BaseType

from fprime_gds.common.utils.string_util import preprocess_fpp_format_str
from fprime_gds.common.loaders import dict_loader
from fprime_gds.common.data_types.exceptions import GdsDictionaryParsingException


PRIMITIVE_TYPE_MAP = {
    "I8": numerical_types.I8Type,
    "I16": numerical_types.I16Type,
    "I32": numerical_types.I32Type,
    "I64": numerical_types.I64Type,
    "U8": numerical_types.U8Type,
    "U16": numerical_types.U16Type,
    "U32": numerical_types.U32Type,
    "U64": numerical_types.U64Type,
    "F32": numerical_types.F32Type,
    "F64": numerical_types.F64Type,
    "bool": BoolType,
}


class JsonLoader(dict_loader.DictLoader):
    """Class to help load JSON dictionaries"""

    # Cache parsed type objects at the class level so they can be reused across subclasses
    parsed_types: dict = {}

    def __init__(self, json_file: str):
        """
        Constructor

        Returns:
            An initialized loader object
        """
        super().__init__()
        self.json_file = json_file
        with open(json_file, "r") as f:
            self.json_dict = json.load(f)

    def get_versions(self):
        """
        Get the framework and project versions of the dictionary

        Returns:
            A tuple of the framework and project versions
        """
        if "metadata" not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Dictionary has no metadata field: {self.json_file}"
            )
        return (
            self.json_dict["metadata"].get("frameworkVersion", "unknown"),
            self.json_dict["metadata"].get("projectVersion", "unknown"),
        )

    def get_metadata(self):
        """Get the metadata field of the JSON dictionary

        Raises:
            GdsDictionaryParsingException: if the dictionary has no metadata field
        """
        if "metadata" not in self.json_dict:
            raise GdsDictionaryParsingException(
                f"Dictionary has no metadata field: {self.json_file}"
            )
        return self.json_dict["metadata"]

    def parse_type(self, type_dict: dict) -> BaseType:
        type_name: str = type_dict.get("name", None)

        if type_name is None:
            raise GdsDictionaryParsingException(
                f"Dictionary entry has no `name` field: {str(type_dict)}"
            )

        if type_name in PRIMITIVE_TYPE_MAP:
            return PRIMITIVE_TYPE_MAP[type_name]

        if type_name == "string":
            return StringType.construct_type(
                f'String_{type_dict["size"]}', type_dict["size"]
            )

        # Check if type has already been parsed
        if type_name in self.parsed_types:
            return self.parsed_types[type_name]

        # Parse new enum/array/serializable types
        qualified_type = None
        for type_def in self.json_dict.get("typeDefinitions", []):
            if type_name == type_def.get("qualifiedName"):
                qualified_type = type_def
                break
        else:
            raise GdsDictionaryParsingException(
                f"Dictionary type name has no corresponding type definition: {type_name}"
            )

        if qualified_type.get("kind") == "alias":
            return self.parse_type(qualified_type.get("underlyingType"))
        
        if qualified_type.get("kind") == "array":
            return self.construct_array_type(type_name, qualified_type)

        if qualified_type.get("kind") == "enum":
            return self.construct_enum_type(type_name, qualified_type)

        if qualified_type.get("kind") == "struct":
            return self.construct_serializable_type(type_name, qualified_type)

        raise GdsDictionaryParsingException(
            f"Dictionary entry has unknown type {str(type_dict)}"
        )

    def construct_enum_type(self, type_name: str, qualified_type: dict) -> EnumType:
        """
        Constructs an EnumType object of the given type name and qualified type dictionary.
        Caches the constructed EnumType object in the parsed_types dictionary.

        Args:
            type_name (str): The name of the enum type.
            qualified_type (dict): A dictionary containing the qualified type information.

        Returns:
            EnumType: The constructed EnumType object.

        """
        enum_dict = {}
        for member in qualified_type.get("enumeratedConstants"):
            enum_dict[member["name"]] = member.get("value")
        enum_type = EnumType.construct_type(
            type_name,
            enum_dict,
            qualified_type["representationType"].get("name"),
        )
        self.parsed_types[type_name] = enum_type
        return enum_type

    def construct_array_type(self, type_name: str, qualified_type: dict) -> ArrayType:
        """
        Constructs an ArrayType object based on the given type name and qualified type dictionary.
        Caches the constructed ArrayType object in the parsed_types dictionary.

        Args:
            type_name (str): The name of the array type.
            qualified_type (dict): The qualified type dictionary containing information about the array type.

        Returns:
            ArrayType: The constructed ArrayType object.

        """
        array_type = ArrayType.construct_type(
            type_name,
            self.parse_type(qualified_type.get("elementType")),
            qualified_type.get("size"),
            JsonLoader.preprocess_format_str(
                qualified_type["elementType"].get("format", "{}")
            ),
        )
        self.parsed_types[type_name] = array_type
        return array_type

    def construct_serializable_type(
        self, type_name: str, qualified_type: dict
    ) -> SerializableType:
        """
        Constructs a SerializableType based on the given type name and qualified type dictionary.
        Caches the constructed SerializableType object in the parsed_types dictionary.

        Args:
            type_name (str): The name of the serializable type.
            qualified_type (dict): The qualified type dictionary containing information about the type.

        Returns:
            SerializableType: The constructed serializable type.

        """
        # Note on struct_members: the order of the members list matter when calling construct_type() below.
        # It should be ordered by incrementing index which corresponds to the order in the FPP declaration
        # The JSON dictionary ordering is not guaranteed, so we use a dict() to sort by index below.
        struct_members = {}
        for name, member_dict in qualified_type.get("members").items():
            member_type_dict = member_dict["type"]
            member_type_obj = self.parse_type(member_type_dict)

            # For member arrays (declared inline, so we create a type on the fly)
            if member_dict.get("size") is not None:
                member_type_obj = ArrayType.construct_type(
                    f"Array_{member_type_obj.__name__}_{member_dict['size']}",
                    member_type_obj,
                    member_dict["size"],
                    JsonLoader.preprocess_format_str(
                        member_dict["type"].get("format", "{}")
                    ),
                )
            fmt_str = JsonLoader.preprocess_format_str(
                member_type_obj.FORMAT if hasattr(member_type_obj, "FORMAT") else "{}"
            )
            description = member_type_dict.get("annotation", "")
            member_index = member_dict["index"]
            if member_index in struct_members:
                raise KeyError(
                    f"Invalid dictionary: Duplicate index {member_index} in serializable type {type_name}"
                )
            struct_members[member_index] = (name, member_type_obj, fmt_str, description)

        # Construct the serializable type with list of members sorted by index
        ser_type = SerializableType.construct_type(
            type_name,
            [struct_members[i] for i in sorted(struct_members.keys())],
        )
        self.parsed_types[type_name] = ser_type
        return ser_type

    @staticmethod
    def preprocess_format_str(format_str: Optional[str]) -> Optional[str]:
        """Preprocess format strings before using them in Python format function
        Internally, this converts FPP-style format strings to Python-style format strings

        Args:
            format_str (str): FPP-style format string

        Returns:
            str: Python-style format string
        """
        if format_str is None:
            return None
        return preprocess_fpp_format_str(format_str)



import argparse
from pathlib import Path
from fprime_gds.common.fpy.bytecode.assembler import assemble, directives_to_fpybc, parse as fpybc_parse
import fprime_gds.common.fpy.error
from fprime_gds.common.fpy.types import deserialize_directives, serialize_directives
import fprime_gds.common.fpy.model
from fprime_gds.common.fpy.model import DirectiveErrorCode, FpySequencerModel
from fprime_gds.common.fpy.parser import parse as fpy_parse
from fprime_gds.common.fpy.codegen import compile

def human_readable_size(size_bytes):
    units = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
    unit_idx = 0
    while size_bytes >= 1024.0 and unit_idx < len(units) - 1:
        size_bytes /= 1024.0
        unit_idx += 1
    size_bytes = int(size_bytes)
    return f"{size_bytes} {units[unit_idx]}"

def compile_main(args: list[str] = None):
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("input", type=Path, help="The input .fpy file")
    arg_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=False,
        default=None,
        help="The output .bin path",
    )
    arg_parser.add_argument(
        "-d",
        "--dictionary",
        type=Path,
        required=True,
        help="The FPrime dictionary .json file",
    )
    arg_parser.add_argument(
        "-b",
        "--bytecode",
        action="store_true",
        default=False,
        help="Whether to output human-readable bytecode instead of binary",
    )
    arg_parser.add_argument(
        "--debug",
        action="store_true",
        default=False,
        help="Pass this to print out compiler debugging information",
    )

    if args is not None:
        args = arg_parser.parse_args(args)
    else:
        args = arg_parser.parse_args()

    if args.debug:
        fprime_gds.common.fpy.error.debug = True

    if not args.input.exists():
        print(f"Input file {args.input} does not exist")
        exit(-1)

    fprime_gds.common.fpy.error.file_name = str(args.input)

    body = fpy_parse(args.input.read_text())
    directives = compile(body, args.dictionary)
    output = args.output
    if output is None:
        if args.bytecode:
            output = args.input.with_suffix(".fpybc")
        else:
            output = args.input.with_suffix(".bin")
    if args.bytecode:
        fpybc = directives_to_fpybc(directives)
        output.write_text(fpybc)
        print(f"{output}")
    else:
        output_bytes, crc = serialize_directives(directives)
        output.write_bytes(output_bytes)
        print(f"{output}\nCRC {hex(crc)} size {human_readable_size(len(output_bytes))}")





def model_main(args: list[str] = None):
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("input", type=Path, help="The input .bin file")
    arg_parser.add_argument(
        "--debug",
        action="store_true",
        help="Whether or not to print debug info during sequence execution",
    )

    if args is not None:
        args = arg_parser.parse_args(args)
    else:
        args = arg_parser.parse_args()

    if not args.input.exists():
        print(f"Input file {args.input} does not exist")
        exit(-1)

    if args.debug:
        fprime_gds.common.fpy.model.debug = True

    directives = deserialize_directives(args.input.read_bytes())
    model = FpySequencerModel()
    ret = model.run(directives)
    if ret != DirectiveErrorCode.NO_ERROR:
        print("Sequence failed with " + str(ret))
        exit(1)


def assemble_main(args: list[str] = None):
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("input", type=Path, help="The input .fpybc file")
    arg_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=False,
        default=None,
        help="The output .bin path",
    )

    if args is not None:
        args = arg_parser.parse_args(args)
    else:
        args = arg_parser.parse_args()

    if not args.input.exists():
        print(f"Input file {args.input} does not exist")
        exit(-1)

    body = fpybc_parse(args.input.read_text())
    directives = assemble(body)
    output = args.output
    if output is None:
        output = args.input.with_suffix(".bin")
    output_bytes, crc = serialize_directives(directives)
    output.write_bytes(output_bytes)
    print(f"{output}\nCRC {hex(crc)} size {human_readable_size(len(output_bytes))}")


def disassemble_main(args: list[str] = None):
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("input", type=Path, help="The input .bin file")
    arg_parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=False,
        default=None,
        help="The output .fpybc path",
    )

    if args is not None:
        args = arg_parser.parse_args(args)
    else:
        args = arg_parser.parse_args()

    if not args.input.exists():
        print(f"Input file {args.input} does not exist")
        exit(-1)

    dirs = deserialize_directives(args.input.read_bytes())
    fpybc = directives_to_fpybc(dirs)
    output = args.output
    if output is None:
        output = args.input.with_suffix(".fpybc")
    output.write_text(fpybc)
    print("Done")
